/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:logging/logging.dart';
import 'package:surec_app/bloc/visual_cubit.dart';

import 'package:surec_app/generated/l10n.dart';
import 'package:surec_app/surecord/surecord_blocrepo.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/ui/routes/route_util.dart';
import 'package:surec_app/ui/theming.dart';
import 'package:surec_app/ui/views/record_list/record_export_list_cubit.dart';


class App extends StatelessWidget {
  final log = Logger("App");

  @override
  Widget build(BuildContext context) {
    log.info('App build');

    return MultiBlocProvider(
      providers: [
        BlocProvider(create: (context) => SuRecordCubit(SuRecordBlocRepo.getInitializedSuRecordBlocRepo())),
        BlocProvider(create: (context) => VisualCubit(VisualState.std())),
        BlocProvider(create: (context) => RecordExportListCubit(),)
      ],      
      child: BlocBuilder<VisualCubit, VisualState>(builder: (context, state) => _buildWithVisualState(context, state)),
    );
  }

  Widget _buildWithVisualState(BuildContext context, VisualState state) {
    return MaterialApp(
        theme: Theming.selectThemeData(context, state.fontSize == FontSize.big, state.contrast == Contrast.high, true),
        darkTheme: Theming.selectThemeData(context, state.fontSize == FontSize.big, state.contrast == Contrast.high, false),
        themeMode : state.mode == Mode.light ? ThemeMode.light : ThemeMode.dark,


    builder: (context, navigator) => _builder(context, navigator, state),

      initialRoute: RouteUtil.START,
      onGenerateRoute: (settings) => RouteUtil.onGenerateRoute(settings),

      localizationsDelegates: [
        S.delegate,
        GlobalMaterialLocalizations.delegate,
        GlobalWidgetsLocalizations.delegate,
        GlobalCupertinoLocalizations.delegate],
      supportedLocales: S.delegate.supportedLocales,


      localeListResolutionCallback: (locales, supportedLocales) {
        log.info("main - locales:");
        log.info(supportedLocales);
        log.info("main - supportedLocales:");
        log.info(supportedLocales);
      },
      localeResolutionCallback: (locale, supportedLocales) {
        log.info("main - locale.countryCode:" + locale.countryCode);
        log.info("main - locale.countryCode:" + locale.languageCode);
        log.info("main - supportedLocales:");
        log.info(supportedLocales);
      },

      debugShowCheckedModeBanner: false,
      debugShowMaterialGrid: false,
    );
  }

  _builder (context, navigator, state) {
    TextTheme ofContextTT =  Theme.of(context).textTheme;
    TextTheme appliedOnOfContextTT = ofContextTT.apply(
      fontSizeDelta: state.fontSize == FontSize.big ? 2 : 0,
      fontSizeFactor: 1.0,
    );

    ColorScheme selectedCS = Theming.selectColorScheme(
        context,
        state.fontSize == FontSize.big, state.contrast == Contrast.high, state.mode == Mode.light);

    ColorScheme ofContextCScopywithSelectedCS = Theme.of(context).colorScheme.copyWith(
      primary: selectedCS.primary,
      primaryVariant: selectedCS.primaryVariant,
      secondary: selectedCS.secondary,
      secondaryVariant: selectedCS.primaryVariant,
      surface: selectedCS.surface,
      background: selectedCS.background,
      error: selectedCS.error,
      onPrimary: selectedCS.onPrimary,
      onSecondary: selectedCS.onSecondary,
      onSurface: selectedCS.onSurface,
      onBackground: selectedCS.onBackground,
      onError: selectedCS.onError,
    );

    ColorScheme constructorCS = ColorScheme(
        primary: selectedCS.primary,
        primaryVariant: selectedCS.primaryVariant,
        secondary: selectedCS.secondary,
        secondaryVariant: selectedCS.primaryVariant,
        surface: selectedCS.surface,
        background: selectedCS.background,
        error: selectedCS.error,
        onPrimary: selectedCS.onPrimary,
        onSecondary: selectedCS.onSecondary,
        onSurface: selectedCS.onSurface,
        onBackground: selectedCS.onBackground,
        onError: selectedCS.onError,
        brightness:  Brightness.light
    );

    print(ofContextCScopywithSelectedCS.primary.value.toRadixString(16));
    print(constructorCS.primary.value.toRadixString(16));

    return Theme(

      data: ThemeData(
        textTheme: appliedOnOfContextTT,
        colorScheme: ofContextCScopywithSelectedCS,
      ),
      child: navigator,
    );
  }



}
