/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as p;
import 'package:surec_app/cache/fs_cache.dart';

import 'fs_cache_exception.dart';


class FSAudioCache extends FSCache {
  final log = Logger('AudioCache');
  static final _tempCaptureItemFileSuffix =".tmp";    // TODO: make adjustable
  static final validFileTypes = [ ".mp3", ".aac", ".mp4"];


  FSAudioCache(String cachePath) : super(cachePath);


  String getTempItemCapturePathWithName() {
    return getTempItemCapturePathWithNameForSuffix(_tempCaptureItemFileSuffix);
  }


  Future<String> cacheItemFileFromPath(String pathWithFilename, String explizitStrippedName) async {
    var file;
    try {
      file = File(pathWithFilename);
    } catch (e, s) {
      final message = "could not open audio file from $pathWithFilename\n$e\n$s";
      log.severe(message);
      return null;
    }

    try {
      await cacheAudioFile(file, explizitStrippedName);
    } catch (e, s) {
      final message = "could not cache audio file from $pathWithFilename\n$e\n$s";
      log.severe(message);
      return null;
    }
  }


  /// returns strippedName of cache entry
  /// Does not throw exception if it fails as consequences would only be cache miss - doesn't break anything
  Future<String> cacheAudioFile(File file, [String explizitStrippedName] ) async {
    var fileAbsolute = file.absolute;
    var path = fileAbsolute.path;
    var basename = p.basename(path);

    var fileType = fileTypeOfBasenameForValidFileTypes(basename, validFileTypes, _tempCaptureItemFileSuffix);
    if (fileType == null) {
      final message = "invalid audio filename $basename";
      log.severe(message);
      return null;
    }

    var baseName;
    var strippedName;

    if( explizitStrippedName != null ) {
      strippedName = explizitStrippedName;
      baseName = explizitStrippedName + fileType;

    } else {
      var fileAbsolute = file.absolute;
      var path = fileAbsolute.path;

      baseName = p.basename(path);
      strippedName = strippedNameForFileTypeAndTmpSuffix(baseName, fileType, _tempCaptureItemFileSuffix);
    }
    var targetPathWithBaseName = p.join(cachePath, baseName);

    try {
      var fileRenamed = file.renameSync(targetPathWithBaseName);
    } catch (e, s) {
      final message = "could not moverename the file $baseName into cache\n$e\n$s";
      log.severe(message);
      return null;
    }
    putIfAbsentIntoKnownCacheItems(strippedName, baseName);
    return Future.value(strippedName);
  }


  Future<ScanAudioCacheFilesResult> scanCacheFiles() async {
    log.info("scanCacheFiles in $cachePath");
    Map<String, String> audios = {};
    try {
      final context = p.Context();
      Directory dir = Directory(cachePath);
      Stream<FileSystemEntity> fileSystemEntityStream = dir.list(recursive: false, followLinks: false);

      await for (FileSystemEntity entity in fileSystemEntityStream) {
        var basename = p.basename(entity.path);
        if ([".jpg", ".png"].contains(context.extension(entity.path, 1).toLowerCase())) {
          audios.putIfAbsent(basename.substring(0, basename.length - 4), () => basename);
          log.info("audio found: " + entity.path);
        } else {
          log.info("scan  miss:" + entity.path);
        }
      }
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "scanForSuRecordFiles reading filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw FSCacheException(FSCacheExceptionCode.FILESYSTEM_ERROR, explanation: message);
    }
    return Future.value(ScanAudioCacheFilesResult(audios));
  }
}


class ScanAudioCacheFilesResult {
  final Map<String, String> audios;

  ScanAudioCacheFilesResult(this.audios);
}


