/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:collection';
import 'dart:io';
import 'dart:typed_data';
import 'package:image/image.dart';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as p;
import 'package:video_thumbnail/video_thumbnail.dart';
import 'fs_cache_exception.dart';


class FSImageAndVideoCache {
  final log = Logger('ImageAndVideoCache');

  final int _thumbnailSize;
  final String _cachePath;
  final bool _useThumbnailMemCache;
  var _thumbnailMemCache = HashMap<String, Uint8List>();

  var _knownThumbnailsInCacheDirectory = Set<String>();
  var _knownImagesInCacheDirectory = Map<String, String>();
  var _knownVideosInCacheDirectory = Map<String, String>();

  final THUMBMAIL_FILETYPE = ".thmb.jpg";
  final THUMBMAIL_FILETYPE_EXTENSIONCOUNT = 2;

  FSImageAndVideoCache(int thumbnailSize, String cachePath, bool useThumbnailMemCache )
      : this._thumbnailSize = thumbnailSize,
        this._cachePath = cachePath,
        this._useThumbnailMemCache = useThumbnailMemCache {
    prepareFSCache(cachePath);
  }

  bool prepareFSCache(String cachePath) {
    var cacheDirectory = Directory(cachePath);
    if (!cacheDirectory.existsSync()) {
      try {
        cacheDirectory.createSync(recursive: true);
      } catch (e, s) {
        final message = "could not create cache dir\n$e\n$s";
        log.severe(message);
        return false;
      }
    }
    return true;
  }

  void initializeKnownCacheItems(knownThumbnailsInCacheDirectory, knownImagesInCacheDirectory, knownVideosInCacheDirectory) {
    _knownThumbnailsInCacheDirectory = knownThumbnailsInCacheDirectory;
    _knownImagesInCacheDirectory = knownImagesInCacheDirectory;
    _knownVideosInCacheDirectory = knownVideosInCacheDirectory;
  }


  String generateThumbnailNameForStrippedName(String strippedName) {
    return "$strippedName$THUMBMAIL_FILETYPE";
  }


  String generateThumbnailPathWithNameForStrippedName(String strippedName) {
    return p.join(_cachePath, "$strippedName$THUMBMAIL_FILETYPE");
  }


  Uint8List getThumbnailForStrippedNameFromMemCache(String strippedName) {
    return _thumbnailMemCache[strippedName];
  }


  String getPathToKnownThumbnailForStrippedName(String strippedName) {
    if (_knownThumbnailsInCacheDirectory.contains(strippedName)) {
      generateThumbnailPathWithNameForStrippedName(strippedName);
    }
  }


  String getPathToKnownImageForStrippedName(String strippedName) {
    if (_knownImagesInCacheDirectory.containsKey(strippedName)) {
      return p.join(_cachePath, _knownImagesInCacheDirectory[strippedName]);
    }
  }


  String getPathToKnownVideoForStrippedName(String strippedName) {
    if (_knownVideosInCacheDirectory.containsKey(strippedName)) {
      return p.join(_cachePath, _knownVideosInCacheDirectory[strippedName]);
    }
  }


  Future<String> cacheImageFileFromPath(String path, String explizitStrippedName) async {
    var file;
    try {
      file = File(path);
    } catch (e, s) {
      final message = "could not open image file from $path\n$e\n$s";
      log.severe(message);
      return null;
    }

    try {
      await cacheImageOrVideoFile(file, true, explizitStrippedName);
    } catch (e, s) {
      final message = "could not cache image file from $path\n$e\n$s";
      log.severe(message);
      return null;
    }
  }


  Future<String> cacheVideoFileFromPath(String path, String explizitStrippedName) async {
    var file;
    try {
      file = File(path);
    } catch (e, s) {
      final message = "could not open video file from $path\n$e\n$s";
      log.severe(message);
      return null;
    }

    try {
      await cacheImageOrVideoFile(file, false, explizitStrippedName);
    } catch (e, s) {
      final message = "could not cache video file from $path\n$e\n$s";
      log.severe(message);
      return null;
    }
  }


  Future<String> cacheImageOrVideoFile(File file, bool isImage, [String explizitStrippedName] ) async {

    var fileAbsolute = file.absolute;
    var path = fileAbsolute.path;

    var basename = p.basename(path);

    if (isImage) {
      if (!(basename.toLowerCase().endsWith(".png") | basename.toLowerCase().endsWith(".jpg"))) {
        final message = "invalid image filename $basename";
        log.severe(message);
        return null;
      }
    } else {
      if (!(basename.toLowerCase().endsWith(".mp4"))) {
        final message = "invalid video filename $basename";
        log.severe(message);
        return null;
      }
    }
    var strippedName = basename.substring(0, basename.length - 4);
    var filetype = basename.substring(basename.length - 4);

    var targetPath = _cachePath;

    Image image;
    Image imageTumbnail;
    Uint8List videoTumbnail;

    if (isImage) {
      image = decodeImage(await File(path).readAsBytes());
      imageTumbnail = copyResizeCropSquare(image, _thumbnailSize);
    } else {
      videoTumbnail = await VideoThumbnail.thumbnailData(
        video: path,
        imageFormat: ImageFormat.JPEG,
        maxWidth: _thumbnailSize,
        quality: 60,
      );
    }

    if (explizitStrippedName != null && explizitStrippedName.length > 0) {
      basename = explizitStrippedName + filetype;
      strippedName = explizitStrippedName;
    }

    var targetPathWithBasename = p.join(_cachePath, basename);
    var thumbnailPathWithName = p.join(_cachePath, "$strippedName$THUMBMAIL_FILETYPE");

    try {
      var fileRenamed = file.renameSync(targetPathWithBasename);
    } catch (e, s) {
      final message = "could not moverename the file into cache\n$e\n$s";
      log.severe(message);
      return null;
    }

    if (isImage) {
      Uint8List thumbnailData;
      try {
        thumbnailData = encodeJpg(imageTumbnail, quality: 60);
        File(thumbnailPathWithName).writeAsBytesSync(thumbnailData, flush: true);
      } catch (e, s) {
        final message = "could not write thumbnail\n$e\n$s";
        log.severe(message);
        return null;
      }
      if (_useThumbnailMemCache) {
        _thumbnailMemCache[strippedName] = thumbnailData;
      }
      _knownImagesInCacheDirectory.putIfAbsent(strippedName, () => basename);
      _knownThumbnailsInCacheDirectory.add(strippedName);
      return Future.value(strippedName);
    } else {
      try {
        File(thumbnailPathWithName).writeAsBytesSync(videoTumbnail);
      } catch (e, s) {
        final message = "could not write thumbnail\n$e\n$s";
        log.severe(message);
        return null;
      }
      if (_useThumbnailMemCache) {
        _thumbnailMemCache[strippedName] = videoTumbnail;
      }
      _knownVideosInCacheDirectory.putIfAbsent(strippedName, () => basename);
      _knownThumbnailsInCacheDirectory.add(strippedName);
      return Future.value(strippedName);
    }
  }


  Future<ScanImageAndVideoCacheFilesResult> scanCacheFiles() async {
    log.info("scanCacheFiles in $_cachePath");
    Set<String> thumbnails = {};
    Map<String, String> images = {};
    Map<String, String> videos = {};
    try {
      final context = p.Context();
      Directory dir = Directory(_cachePath);
      Stream<FileSystemEntity> fileSystemEntityStream = dir.list(recursive: false, followLinks: false);

      await for (FileSystemEntity entity in fileSystemEntityStream) {
        var basename = p.basename(entity.path);
        if (context.extension(entity.path, THUMBMAIL_FILETYPE_EXTENSIONCOUNT).toLowerCase() == THUMBMAIL_FILETYPE) {
          thumbnails.add(basename.substring(0, basename.length - THUMBMAIL_FILETYPE.length));
          log.info("thumbnail found: " + entity.path);
        } else if ([".jpg", ".png"].contains(context.extension(entity.path, 1).toLowerCase())) {
          images.putIfAbsent(basename.substring(0, basename.length - 4), () => basename);
          log.info("image found: " + entity.path);
        } else if ([".mp4"].contains(context.extension(entity.path, 1).toLowerCase())) {
          videos.putIfAbsent(basename.substring(0, basename.length - 4), () => basename);
          log.info("video found: " + entity.path);
        } else {
          log.info("scan  miss:" + entity.path);
        }
      }
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "scanForSuRecordFiles reading filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw FSCacheException(FSCacheExceptionCode.FILESYSTEM_ERROR, explanation: message);
    }
    return Future.value(ScanImageAndVideoCacheFilesResult(thumbnails, images, videos));
  }


  // return the set of names of cached thumbnails
  Set<String> loadThumbnailsFromFsIntoMemCache(Set<String> thumbnailStrippedNames) {
    if(_useThumbnailMemCache) {
      var cachedToMem = Set<String>();
      for (var thumbnailStrippedName in thumbnailStrippedNames) {
        var thumbnailPath;
        var thumbnailFile;
        try {
          thumbnailPath = generateThumbnailPathWithNameForStrippedName(thumbnailStrippedName);
          thumbnailFile = File(thumbnailPath);
          Uint8List thumbnailImageData = thumbnailFile.readAsBytesSync();
          _thumbnailMemCache[thumbnailStrippedName] = thumbnailImageData;
          cachedToMem.add(thumbnailStrippedName);
        } catch (exception, stacktrace) {
          final exceptionStr = exception.toString();
          final stacktraceStr = stacktrace.toString();
          final message = "loadThumbnailsIntoMem failed for $thumbnailFile at thumbnailPath";
          log.severe("$message\n$exceptionStr\n$stacktraceStr");
        }
      }
      return cachedToMem;
    } else {
      return Set<String>();
    }
  }


  HashMap<String, Uint8List> resetThumbnailMemCache() {
    _thumbnailMemCache = HashMap<String, Uint8List>();
    return _thumbnailMemCache;
  }
}


class ScanImageAndVideoCacheFilesResult {
  final Set<String> thumbnails;
  final Map<String, String> images;
  final Map<String, String> videos;

  ScanImageAndVideoCacheFilesResult(this.thumbnails, this.images, this.videos);
}


