/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:collection';
import 'dart:io';
import 'dart:math';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as p;
import 'fs_cache_exception.dart';


class FSNoteCache {
  final log = Logger('FSNoteCache');

  final String _cachePath;
  final int _textsnippetSize;
  final bool _useTextsnippetMemCache;
  var _textsnippetMemCache = HashMap<String, String>();

  var _knownTextsnippetsInCacheDirectory = Set<String>();
  var _knownTextsInCacheDirectory = Map<String, String>();
  final TEXTSNIPPET_FILETYPE = ".thmb.txt";
  final TEXTSNIPPET_FILETYPE_EXTENSIONCOUNT = 2;

  FSNoteCache(int textsnippetSize, String cachePath, bool useTextsnippetMemCache)
      : this._textsnippetSize = textsnippetSize,
        this._cachePath = cachePath,
        _useTextsnippetMemCache = useTextsnippetMemCache {
    prepareFSCache(cachePath);
  }

  bool prepareFSCache(String cachePath) {
    var cacheDirectory = Directory(cachePath);
    if (!cacheDirectory.existsSync()) {
      try {
        cacheDirectory.createSync(recursive: true);
      } catch (e, s) {
        final message = "could not create cache dir\n$e\n$s";
        log.severe(message);
        return false;
      }
    }
    return true;
  }

  void initializeKnownCacheItems(knownTextsnippetsInCacheDirectory, knownTextsInCacheDirectory, ) {
    _knownTextsnippetsInCacheDirectory = knownTextsnippetsInCacheDirectory;
    _knownTextsInCacheDirectory = knownTextsInCacheDirectory;
  }


  String generateTextsnippetNameForStrippedName(String strippedName) {
    return "$strippedName$TEXTSNIPPET_FILETYPE";
  }


  String generateTextsnippetPathWithNameForStrippedName(String strippedName) {
    return p.join(_cachePath, "$strippedName$TEXTSNIPPET_FILETYPE");
  }


  String getTextsnippetForStrippedNameFromMemCache(String strippedName) {
    return _textsnippetMemCache[strippedName];
  }


  String getPathToKnownTextsnippetForStrippedName(String strippedName) {
    if (_knownTextsnippetsInCacheDirectory.contains(strippedName)) {
      return generateTextsnippetPathWithNameForStrippedName(strippedName);
    }
  }


  String getPathToKnownTextForStrippedName(String strippedName) {
    if (_knownTextsInCacheDirectory.containsKey(strippedName)) {
      return p.join(_cachePath, _knownTextsInCacheDirectory[strippedName]);
    }
  }


  Future<String> cacheText(String text, String strippedName) async {

    if (strippedName == null || strippedName.length == 0)  {
      final message = "invalid strippedName";
      log.severe(message);
      return null;
    }

    var targetPath = _cachePath;
    var filetype = ".txt";
    var basename = strippedName + filetype;
    var targetPathWithBasename = p.join(_cachePath, basename);
    var textsnippetPathWithName = p.join(_cachePath, "$strippedName$TEXTSNIPPET_FILETYPE");

    try {
      var targetFile = File(targetPathWithBasename);
      targetFile.writeAsStringSync(text);
    } catch (e, s) {
      final message = "could not create the file in cache\n$e\n$s";
      log.severe(message);
      return null;
    }

    var textsnippet = text.substring(0, min(_textsnippetSize, text.length));
    try {
      File(textsnippetPathWithName).writeAsString(textsnippet);
    } catch (e, s) {
      final message = "could not create the textsnippet file in cache\n$e\n$s";
      log.severe(message);
      return null;
    }
    if(_useTextsnippetMemCache) {
      _textsnippetMemCache[strippedName] = textsnippet;
    }
    _knownTextsInCacheDirectory.putIfAbsent(strippedName, () => basename);
    _knownTextsnippetsInCacheDirectory.add(strippedName);
    return Future.value(strippedName);
  }


  Future<ScanNoteCacheFilesResult> scanCacheFiles() async {
    log.info("scanCacheFiles in $_cachePath");
    Set<String> textsnippets = {};
    Map<String, String> texts = {};
    try {
      final context = p.Context();
      Directory dir = Directory(_cachePath);
      Stream<FileSystemEntity> fileSystemEntityStream = dir.list(recursive: false, followLinks: false);

      await for (FileSystemEntity entity in fileSystemEntityStream) {
        var basename = p.basename(entity.path);
        if (context.extension(entity.path, TEXTSNIPPET_FILETYPE_EXTENSIONCOUNT).toLowerCase() == TEXTSNIPPET_FILETYPE) {
          textsnippets.add(basename.substring(0, basename.length - TEXTSNIPPET_FILETYPE.length));
          log.info("textsnippet found: " + entity.path);
        } else if ([".txt"].contains(context.extension(entity.path, 1).toLowerCase())) {
          texts.putIfAbsent(basename.substring(0, basename.length - 4), () => basename);
          log.info("text found: " + entity.path);
        } else {
          log.info("scan  miss:" + entity.path);
        }
      }
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "scanCacheFiles reading filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw FSCacheException(FSCacheExceptionCode.FILESYSTEM_ERROR, explanation: message);
    }
    return Future.value(ScanNoteCacheFilesResult(textsnippets, texts));
  }


  Set<String> loadTextsnippetsFromFsIntoMemCache(Set<String> textsnippetsStrippedNames) {
    if(_useTextsnippetMemCache) {
      var cachedToMem = Set<String>();
      for (var textsnippetStrippedName in textsnippetsStrippedNames) {
        var textsnippetPath;
        var textsnippetFile;
        try {
          textsnippetPath = generateTextsnippetPathWithNameForStrippedName(textsnippetStrippedName);
          textsnippetFile = File(textsnippetPath);
          String textsnippetData = textsnippetFile.readAsStringSync();
          _textsnippetMemCache[textsnippetStrippedName] = textsnippetData;
          cachedToMem.add(textsnippetStrippedName);
        } catch (exception, stacktrace) {
          final exceptionStr = exception.toString();
          final stacktraceStr = stacktrace.toString();
          final message = "loadThumbnailsIntoMem failed for $textsnippetFile at thumbnailPath";
          log.severe("$message\n$exceptionStr\n$stacktraceStr");
        }
      }
      return cachedToMem;
    } else {
      return Set<String>();
    }
  }

  HashMap<String, String> resetTextsnippetMemCache() {
    _textsnippetMemCache = HashMap<String, String>();
    return _textsnippetMemCache;
  }
}

class ScanNoteCacheFilesResult {
  final Set<String> textsnippets;
  final Map<String, String> texts;

  ScanNoteCacheFilesResult(this.textsnippets, this.texts);
}


