/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:equatable/equatable.dart';
import 'package:surec_app/surecord/surecord_meta.dart';

abstract class Filter extends Equatable {
  Filter(Matcher matcher);
}

abstract class Matcher extends Equatable {
  Matcher(BaseAdapter baseAdapter, FilterExpression filterExpression);
}

abstract class FilterExpression extends Equatable {
  bool match(BaseAdapter baseAdapter);
}

// ------------------------------

class AndExpression extends FilterExpression {
  FilterExpression _leftFilterExpression;
  FilterExpression _rightFilterExpression;
  AndExpression(FilterExpression this._leftFilterExpression, FilterExpression this._rightFilterExpression);

  @override
  bool match(BaseAdapter baseAdapter) {
    return _leftFilterExpression.match(baseAdapter) && _rightFilterExpression.match(baseAdapter);
  }

  @override
  List<Object> get props => [_leftFilterExpression, _rightFilterExpression];
}

class OrExpression extends FilterExpression {
  FilterExpression _leftFilterExpression;
  FilterExpression _rightFilterExpression;
  OrExpression(FilterExpression this._leftFilterExpression, FilterExpression this._rightFilterExpression);

  @override
  bool match(BaseAdapter baseAdapter) {
    return _leftFilterExpression.match(baseAdapter) || _rightFilterExpression.match(baseAdapter);
  }

  @override
  List<Object> get props => [_leftFilterExpression, _rightFilterExpression];
}

class NotExpression extends FilterExpression {
  final FilterExpression _filterExpression;
  NotExpression(this._filterExpression);

  @override
  bool match(BaseAdapter baseAdapter) {
    return ! _filterExpression.match(baseAdapter);
  }

  @override
  List<Object> get props => [_filterExpression];
}


class IdExpression extends FilterExpression {
  final FilterExpression _filterExpression;
  IdExpression(this._filterExpression);

  @override
  bool match(BaseAdapter baseAdapter) {
    return _filterExpression.match(baseAdapter);
  }

  @override
  List<Object> get props => [_filterExpression];
}


class AttributeValueStringListContainsExpression extends FilterExpression {
  final BaseAdapterAttribute _adapterAttribute;
  final String _value;
  AttributeValueStringListContainsExpression(this._adapterAttribute, this._value);

  bool match(BaseAdapter baseAdapter) {
    List<String> sl= baseAdapter.getStringListAttribute(_adapterAttribute);
    return sl.contains(_value);
  }

  @override
  List<Object> get props => [_adapterAttribute, _value];
}

class AttributeValueStringListNotContainsExpression extends FilterExpression {
  final BaseAdapterAttribute _adapterAttribute;
  final String _value;
  AttributeValueStringListNotContainsExpression(this._adapterAttribute, this._value);

  bool match(BaseAdapter baseAdapter) {
    var expression = AttributeValueStringListContainsExpression(this._adapterAttribute, this._value);
    return !expression.match(baseAdapter);
  }

  @override
  List<Object> get props => [_adapterAttribute, _value];
}

class AttributeValueStringIsExpression extends FilterExpression {
  final BaseAdapterAttribute _adapterAttribute;
  final String _value;
  AttributeValueStringIsExpression(this._adapterAttribute, this._value);

  bool match(BaseAdapter baseAdapter) {
    String sl= baseAdapter.getStringAttribute(_adapterAttribute);
    // TODO: implement when necessary
    throw(Exception("niy"));
  }

  @override
  List<Object> get props => [_adapterAttribute, _value];
}

class AttributeValueIntIsLessOrEqual extends FilterExpression {
  final BaseAdapterAttribute _adapterAttribute;
  final int _value;
  AttributeValueIntIsLessOrEqual(this._adapterAttribute, this._value);

  bool match(BaseAdapter baseAdapter) {
    return baseAdapter.getIntAttribute(_adapterAttribute) <= _value;
  }

  @override
  List<Object> get props => [_adapterAttribute, _value];
}

class AttributeValueIntIsGreaterOrEqual extends FilterExpression {
  final BaseAdapterAttribute _adapterAttribute;
  final int _value;
  AttributeValueIntIsGreaterOrEqual(this._adapterAttribute, this._value);

  bool match(BaseAdapter baseAdapter) {
    return baseAdapter.getIntAttribute(_adapterAttribute) >= _value;
  }

  @override
  List<Object> get props => [_adapterAttribute, _value];
}
// --------------------------------------------------------------

abstract class BaseAdapterType {
  const BaseAdapterType._internal();
}

class RecordAdapterType implements BaseAdapterType {
  final _value;
  const RecordAdapterType._internal(this._value);
  toString() => 'RecordAdapterType.$_value';

  static const bool = const RecordAdapterType._internal('bool');
  static const int = const RecordAdapterType._internal('int');
  static const double = const RecordAdapterType._internal('double');
  static const string = const RecordAdapterType._internal('string');
  static const stringList = const RecordAdapterType._internal('stringList');
}

// --------------------------------------------------------------

abstract class BaseAdapterAttribute {
  const BaseAdapterAttribute._internal();
}

abstract class BaseAdapter extends Equatable {
  getIntAttribute(BaseAdapterAttribute attr);
  getDoubleAttribute(BaseAdapterAttribute attr);
  getStringAttribute(BaseAdapterAttribute attr);
  getStringListAttribute(BaseAdapterAttribute attr);
}






