/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';
import 'dart:typed_data';

import 'package:logging/logging.dart';
import 'package:surec_app/filter/filter.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_exception.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_repo.dart';
import 'package:surec_app/surecord/surecord_translate.dart';
import 'package:surec_app/surecord/surecord_util.dart';
import 'package:synchronized/synchronized.dart';


class SaveResult {
  bool   ok;
  String id_sha3_512;
  List<BlocSuRecord> blocSuRecords;

  SaveResult(this.ok, this.id_sha3_512, this.blocSuRecords);
}


class DeleteResult {
  bool   ok;
  List<BlocSuRecord> blocSuRecords;
  DeleteResult(this.ok, this.blocSuRecords);
}


class ExistsResult {
  bool   ok;
  bool   exists;

  ExistsResult(this.ok, this.exists);
}


class FindResult {
  bool   ok;
  List<BlocSuRecord> blocSuRecords;

  FindResult(this.ok, this.blocSuRecords);
}


class CountResult {
  bool   ok;
  int count;

  CountResult(this.ok, this.count);
}


class StatisticsResult {
  bool   ok;
  int count;
  Set<String> tags;
  Set<String> refs;

  StatisticsResult(this.ok, this.count, this.tags, this.refs);
}


class SuRecordBlocRepo {
  final log = Logger('SuRecordBlocRepo');


  SuRecordBlocRepo._internal();
  static final SuRecordBlocRepo _suRecordBlocRepo = SuRecordBlocRepo._internal();


  static SuRecordFilesRepo _suRecordFilesRepo;
  static List<BlocSuRecord> _blocSuRecords;
  static final _blocRepoLock = new Lock();


  static Future<void> _initializedBlocSuRecords;
  Future<void> get initialized => _initializedBlocSuRecords;

  static getInitializedSuRecordBlocRepo() {
    return _suRecordBlocRepo;
  }

  factory SuRecordBlocRepo(SuRecordFilesRepo suRecordFilesRepo) {
    _suRecordFilesRepo = suRecordFilesRepo;
    _initializedBlocSuRecords = _updateBlocSuRecords();
    return _suRecordBlocRepo;
  }


  static Future<void> _updateBlocSuRecords() async {
    List<SuRecordFile> suRecordFiles = await _suRecordFilesRepo.findAll();
    _blocSuRecords = SuRecordTranslate.toBlocSuRecordsfromSuRecordFiles(suRecordFiles);
  }


  Future<SaveResult> saveFile(
      String pathWithFilename, SuRecordType suRecordType, List<String> tags, List<String> refs) async {
    return await _blocRepoLock.synchronized(() async {
      try {

        File dataFile = new File(pathWithFilename);
        List<int> bytes = dataFile.readAsBytesSync();

        SuRecordFile suRecordFile = await _suRecordFilesRepo.save(
            data: bytes, tags: tags, refs: refs, suRecordType: suRecordType, suRecordFile: null);
        await _updateBlocSuRecords();

        final String id_sha3_512 = SuRecordUtil.idHash_fromSuRecordFile(suRecordFile);

        return Future.value(SaveResult(true, id_sha3_512, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("saveFile",e,s);
        return Future.value(SaveResult(false, "", []));
      }
    });
  }


  /// changes and returns new _blocSuRecords list
  ///
  /// Hints:
  /// create the data param like: ``` Uint8List data = utf8.encode("Some Note"); ``` for a note text
  Future<SaveResult> saveMem(
      Uint8List data, SuRecordType suRecordType, List<String> tags, List<String> refs) async {
    return await _blocRepoLock.synchronized(() async {
      try {

        SuRecordFile suRecordFile = await _suRecordFilesRepo.save(
            data: data, tags: tags, refs: refs, suRecordType: suRecordType, suRecordFile: null);
        await _updateBlocSuRecords();

        final String id_sha3_512 = SuRecordUtil.idHash_fromSuRecordFile(suRecordFile);

        return Future.value(SaveResult(true, id_sha3_512, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("saveMem",e,s);
        return Future.value(SaveResult(false, "", []));
      }
    });
  }


  /// changes and returns new _blocSuRecords list
  Future<SaveResult> saveMetaTagsRefs(String id_sha3_512, List<String> tags, List<String> refs) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        SuRecordFile suRecordFile = await _suRecordFilesRepo.findByID(id_sha3_512);

        SuRecordFile suRecordFileResult = await _suRecordFilesRepo.save(
            data: null,
            tags: tags,
            refs: refs,
            verifications: suRecordFile.suRecord.verifications,
            suRecordType: suRecordFile.suRecord.type,
            suRecordFile: suRecordFile);
        await _updateBlocSuRecords();

        return Future.value(SaveResult(true, id_sha3_512, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("saveMetaTagsRefs", e, s);
        return Future.value(SaveResult(false, "", []));
      }
    });
  }


  /// changes list -> returns new list
  Future<SaveResult> addVerification(String id_sha3_512, Verification verification) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        SuRecordFile suRecordFile = await _suRecordFilesRepo.findByID(id_sha3_512);

        List<Verification> increasedVerifications =
          new List<Verification>.from(suRecordFile.suRecord.verifications)
            ..add(verification);

        SuRecordFile suRecordFileResult = await _suRecordFilesRepo.save(
            data: null,
            tags: suRecordFile.suRecord.tags,
            refs: suRecordFile.suRecord.refs,
            verifications: increasedVerifications,
            suRecordType: suRecordFile.suRecord.type,
            suRecordFile: suRecordFile);
        await _updateBlocSuRecords();

        return Future.value(SaveResult(true, id_sha3_512, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("addVerification",e,s);
        return Future.value(SaveResult(false, "", []));
      }
    });
  }

  /// changes list -> returns new list
  Future<SaveResult> saveVerifications(String id_sha3_512, List<Verification> verifications) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        SuRecordFile suRecordFile = await _suRecordFilesRepo.findByID(id_sha3_512);

        SuRecordFile suRecordFileResult = await _suRecordFilesRepo.save(
            data: null,
            tags: suRecordFile.suRecord.tags,
            refs: suRecordFile.suRecord.refs,
            verifications: verifications,
            suRecordType: suRecordFile.suRecord.type,
            suRecordFile: suRecordFile);
        await _updateBlocSuRecords();

        return Future.value(SaveResult(true, id_sha3_512, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("saveVerification",e,s);
        return Future.value(SaveResult(false, "", []));
      }
    });
  }


  /// changes list -> returns list
  Future<DeleteResult> delete(String id) async {
    return await _blocRepoLock.synchronized(() async {
      try {

        bool deleted = await _suRecordFilesRepo.delete(id);
        await _updateBlocSuRecords();

        return Future.value(DeleteResult(true, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("delete", e, s);
        return Future.value(DeleteResult(false, []));
      }
    });
  }


  /// does not change list -> returns ExistsResult: OK-State and info if found.
  Future<ExistsResult> exists(String id_sha3_512) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        SuRecordFile suRecordFile = await _suRecordFilesRepo.findByID(id_sha3_512);

        return Future.value(ExistsResult(true, true));
      } catch (e, s) {
        log.info("exists", e, s);
        return Future.value(ExistsResult(false, false));
      }
    });
  }


  /// does not change list -> returns FindResult: OK-State and single Element blocSuRecord list
  Future<FindResult> findById(String id_sha3_512) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        SuRecordFile suRecordFile = await _suRecordFilesRepo.findByID(id_sha3_512);

        final BlocSuRecord blocSuRecord = SuRecordTranslate.toBlocSuRecord(suRecordFile.suRecord);

        return Future.value(FindResult(true, [blocSuRecord]));
      } catch (e, s) {
        log.info("findById", e, s);
        return Future.value(FindResult(false, []));
      }
    });
  }


  /// does not change list -> returns FindResult: OK-State and blocSuRecord list
  Future<FindResult> findAll() async {
    return await _blocRepoLock.synchronized(() async {
      try {
        final List<SuRecordFile> allSuRecordFiles = await _suRecordFilesRepo.findAll();
        final List<BlocSuRecord> blocSuRecords = SuRecordTranslate.toBlocSuRecordsfromSuRecordFiles(allSuRecordFiles);

        return Future.value(FindResult(true, List.unmodifiable(_blocSuRecords)));
      } catch (e, s) {
        log.info("findAll", e, s);
        return Future.value(FindResult(false, []));
      }
    });
  }


  /// does not change list -> returns FindResult: OK-State and blocSuRecord list
  Future<void> findByFilter(FilterExpression filterExpression) async {
    return await _blocRepoLock.synchronized(() async {
      try {
        throw SuRecordException(SuRecordExceptionCode.NOT_IMPLEMENTED);

        // TODO: implement when necessary

      } catch (e, s) {
        log.info("findByFilter", e, s);
        return Future.value(FindResult(false, []));
      }

    });
  }


  /// does not change list -> returns StatisticsResult: OK-State and additional info
  Future<StatisticsResult> statistics() async {
    return await _blocRepoLock.synchronized(() async {
      try {
        final SuRecordFilesRepoStatistics suRecordFilesRepoCountResult = await _suRecordFilesRepo.statistics();

        return Future.value(StatisticsResult(true,
            suRecordFilesRepoCountResult.count,
            suRecordFilesRepoCountResult.tags,
            suRecordFilesRepoCountResult.refs));
      } catch (e, s) {
        log.info("statistics",e,s);
        return Future.value(StatisticsResult(false, 0, Set<String>(), Set<String>()));
      }
    });
  }


  /// does not change list -> returns CountResult: OK-State and additional info
  Future<CountResult> count() async {
    return await _blocRepoLock.synchronized(() async {
      try {
        // TODO: optimize, don't handle tags & refs
        final SuRecordFilesRepoStatistics suRecordFilesRepoCountResult = await _suRecordFilesRepo.statistics();

        return Future.value(CountResult(true, suRecordFilesRepoCountResult.count));
      } catch (e,s) {
        log.info("count",e,s);
        return Future.value(CountResult(false, 0));
      }
    });
  }
}
