/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';
import 'dart:typed_data';

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:logging/logging.dart';
import 'package:surec_app/filter/filter.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_blocrepo.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_repo.dart';
import 'package:surec_app/surecord/surecord_translate.dart';
import 'package:surec_app/surecord/surecord_util.dart';
import 'package:synchronized/synchronized.dart';

part 'surecord_state.dart';

const String NOT_IMPLEMENTED_YET = "not implemented";


class SuRecordCubit extends Cubit<SuRecordCubitState> {
  final log = Logger('SuRecordCubit');

  SuRecordBlocRepo _suRecordBlocRepo;

  SuRecordCubit(SuRecordBlocRepo suRecordBlocRepo) : super(SuRecordInitialCubitState()) {
    this._suRecordBlocRepo = suRecordBlocRepo;
  }


  /// changes list -> returns new list
  ///
  /// Hints:
  /// The [pathWithName] has to be usable be usable by code like
  /// return File(pathWithName);
  Future<void> saveFile(
      String pathWithName,
      SuRecordType suRecordType,
      List<String> tags,
      List<String> refs,
      [String transaction=""]) async {

    emit(SuRecordWorkingCubitState());

    try {
      File f = new File(pathWithName);
      Uint8List data = await f.readAsBytes();

      SaveResult saveResult = await _suRecordBlocRepo.saveMem(data, suRecordType, tags, refs);

      if (saveResult.ok) {
        emit(SuRecordSavedCubitState(saveResult.id_sha3_512, saveResult.blocSuRecords, transaction));
      } else {
        emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
      }
    } catch (exception, stackTrace) {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));

      final exceptionStr = exception.toString();
      final stackTraceStr = stackTrace.toString();
      final message = "saveFile failed";
      log.severe("$message\n$exceptionStr\n$stackTraceStr");
    }
  }


  /// changes list -> returns new list
  ///
  /// Hints:
  /// create the data param like: ``` Uint8List data = utf8.encode("Some Note"); ``` for a note text
  Future<void> saveMem(
      Uint8List data, SuRecordType suRecordType, List<String> tags, List<String> refs, [String transaction=""])
  async {
    emit(SuRecordWorkingCubitState());

    SaveResult saveResult = await _suRecordBlocRepo.saveMem(
        data, suRecordType, tags, refs);

    if(saveResult.ok) {
      emit(SuRecordSavedCubitState(saveResult.id_sha3_512, saveResult.blocSuRecords, transaction));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
    }
  }

  /// changes list -> returns list
  Future<void> saveMetaTagsRefs(String id_sha3_512, List<String> tags, List<String> refs, [String transaction=""])
  async {
    emit(SuRecordWorkingCubitState());

    SaveResult saveResult = await _suRecordBlocRepo.saveMetaTagsRefs(id_sha3_512, tags, refs);

    if(saveResult.ok) {
      emit(SuRecordSavedCubitState(id_sha3_512, saveResult.blocSuRecords, transaction));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
    }
  }

  /// changes list -> returns list
  Future<bool> delete(String id, [String transaction=""]) async {
    emit(SuRecordWorkingCubitState());

    DeleteResult deleteResult = await _suRecordBlocRepo.delete(id);

    if(deleteResult.ok) {
      emit(SuRecordDeletedCubitState(true, id, deleteResult.blocSuRecords));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
    }
  }

  /// bool oder single oder single as list
  Future<void> exists(String id, [String transaction=""]) async {
    emit(SuRecordWorkingCubitState());

    ExistsResult existsResult = await _suRecordBlocRepo.exists(id);

    if(existsResult.ok) {
      emit(SuRecordExistsCubitState(existsResult.exists, id));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
    }
  }

  // does not change list -> returns single
  Future<void> findById(String id, [String transaction=""]) async {
    emit(SuRecordWorkingCubitState());
    final found = false;
    final SuRecord suRecord = null; // await _suRecordRepo.fetch(param)
    // TODO: implement

    FindResult findResult = await _suRecordBlocRepo.findById(id);

    if(findResult.ok) {
      SuRecordFindByIDCubitState suRecordSingleCubitState =
      SuRecordFindByIDCubitState(found, id, findResult.blocSuRecords.first, transaction);
      emit(suRecordSingleCubitState);
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED, transaction));
    }
  }

  // asks for list -> returns list
  Future<void> findAll([String transaction=""]) async {
    emit(SuRecordWorkingCubitState());
    FindResult findResult = await _suRecordBlocRepo.findAll();

    if(findResult.ok) {
      emit(SuRecordFindByFilterCubitState(null, findResult.blocSuRecords, transaction));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED));
    }
  }

  // asks for list -> returns list
  Future<void> findByFilter(FilterExpression filterExpression, [String transaction=""]) async {
    emit(SuRecordWorkingCubitState());

    // TODO: implement real filtering based on SuRecordFilter suRecordFilter
    FindResult findResult = await _suRecordBlocRepo.findAll();

    if(findResult.ok) {
      emit(SuRecordFindByFilterCubitState(null, findResult.blocSuRecords, transaction));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED));
    }
  }

  // int oder extended info (tags)
  Future<void> count([String transaction=""]) async {
    emit(SuRecordWorkingCubitState());

    StatisticsResult statisticsResult = await _suRecordBlocRepo.statistics();

    if(statisticsResult.ok) {
      emit(SuRecordCountCubitState(statisticsResult.count, statisticsResult.tags, transaction));
    } else {
      emit(SuRecordFailureCubitState(SuRecordFailureCubitStateValue.UNSPECIFIED));
    }
  }
}
