/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:json_annotation/json_annotation.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
part 'surecord_meta.g.dart';

String algoToSring(Algo algo) {
  return _$AlgoEnumMap[algo];
}

String suRecordTypeToString(SuRecordType suRecordFileType) {
  return _$SuRecordTypeEnumMap[suRecordFileType];
}

@JsonSerializable(explicitToJson: true)
class Fingerprint {
  @JsonKey(disallowNullValue: true, required: true)
  Algo algo;  // enum

  @JsonKey(disallowNullValue: true, required: true)
  String value;   //  hexstring

  Fingerprint(this.algo, this.value );

  factory Fingerprint.fromJson(Map<String, dynamic> json) => _$FingerprintFromJson(json);
  Map<String, dynamic> toJson() => _$FingerprintToJson(this);
}

@JsonSerializable(explicitToJson: true)
class Fingerprints {
  @JsonKey(disallowNullValue: true, required: true)
  List<Fingerprint> fingerprints;

  Fingerprints(this.fingerprints );

  factory Fingerprints.fromJson(Map<String, dynamic> json) => _$FingerprintsFromJson(json);
  Map<String, dynamic> toJson() => _$FingerprintsToJson(this);
}


@JsonSerializable(explicitToJson: true)
class Verification {
  @JsonKey(disallowNullValue: true, required: true)
  String uri; // of verifier

  @JsonKey(disallowNullValue: true, required: true)
  int timestamp; // 32 bit utc system time
  // use this for sort by time with "local" uri

  @JsonKey(disallowNullValue: true, required: true)
  List<Fingerprint> fingerprints;

  @JsonKey(disallowNullValue: true, required: true)
  Confirmation confirmation;

  @JsonKey(disallowNullValue: true, required: true)
  Verification(this.uri, this.timestamp, this.fingerprints, this.confirmation);

  factory Verification.fromJson(Map<String, dynamic> json) => _$VerificationFromJson(json);
  Map<String, dynamic> toJson() => _$VerificationToJson(this);
}


@JsonSerializable()
class Confirmation {
  @JsonKey(disallowNullValue: true, required: true)
  String value;

  @JsonKey(disallowNullValue: true, required: true)
  Algo algo;

  @JsonKey(disallowNullValue: true, required: true)
  int timestamp; // 32 bit utc system time -> creation modification

  Confirmation(this.value, this.algo, this.timestamp);

  factory Confirmation.fromJson(Map<String, dynamic> json) => _$ConfirmationFromJson(json);
  Map<String, dynamic> toJson() => _$ConfirmationToJson(this);
}


///Implements the SuRecord DAO containing MetaData for the verifiable file
///
/// @param verifications
/// @param filename of the file that will be verifiable and that can not be changed
/// @param type
/// @param tags
/// @param refs
/// @param modificationTime of the metadata
@JsonSerializable(explicitToJson: true)
class SuRecord {

  @JsonKey(disallowNullValue: true, required: true)
  List<Verification> verifications;

  @JsonKey(disallowNullValue: true, required: true)
  String dataFilename;

  @JsonKey(disallowNullValue: true, required: true)
  SuRecordType type;      // is redundant (filename suffix) -> maybe better remove  // TODO: discuss

  @JsonKey(disallowNullValue: true, required: true)
  List<String> tags;  // user defined tag strings

  @JsonKey(disallowNullValue: true, required: true)
  List<String> refs;  // format: hex-strings. length may vary when different digest algos are used

  @JsonKey(disallowNullValue: true, required: true)
  int modificationTime; // 32 bit utc system time -> time of latest modification

  SuRecord(this.verifications, this.tags, this.refs, this.modificationTime, this.type, this.dataFilename);

  factory SuRecord.fromJson(Map<String, dynamic> json) => _$SuRecordFromJson(json);
  Map<String, dynamic> toJson() => _$SuRecordToJson(this);
}


///Implements a holder for SuRecord DAO together with it filename on local storage
///
/// @param suRecord the metadata DAO
/// @param filename of the .surec file containing the metadata DAO and the verifiable file
@JsonSerializable(explicitToJson: true)
class SuRecordFile {

  @JsonKey(disallowNullValue: true, required: true)
  SuRecord suRecord;

  @JsonKey(disallowNullValue: true, required: true)
  String surecFilename;

  SuRecordFile(this.suRecord, this.surecFilename);

  factory SuRecordFile.fromJson(Map<String, dynamic> json) => _$SuRecordFileFromJson(json);
  Map<String, dynamic> toJson() => _$SuRecordFileToJson(this);
}


@JsonSerializable()
class SuRecords {
  @JsonKey(disallowNullValue: true, required: true)
  List<SuRecord> suRecords;

  @JsonKey(disallowNullValue: true, required: true)
  int modificationTime; // 32 bit utc system time -> latest modification

  SuRecords(this.suRecords, this.modificationTime);

  factory SuRecords.fromJson(Map<String, dynamic> json) => _$SuRecordsFromJson(json);
  Map<String, dynamic> toJson() => _$SuRecordsToJson(this);
}


@JsonSerializable()
class SuRecordFiles {
  @JsonKey(disallowNullValue: true, required: true)
  List<SuRecordFile> suRecordFiles;

  @JsonKey(disallowNullValue: true, required: true)
  int modificationTime; // 32 bit utc system time -> latest modification

  SuRecordFiles(this.suRecordFiles, this.modificationTime);

  factory SuRecordFiles.fromJson(Map<String, dynamic> json) => _$SuRecordFilesFromJson(json);
  Map<String, dynamic> toJson() => _$SuRecordFilesToJson(this);
}
