/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:async';
import 'dart:io';
import 'dart:typed_data'; // see https://api.dart.dev/stable/2.10.3/dart-typed_data/dart-typed_data-library.html

import 'package:equatable/equatable.dart';
import 'package:logging/logging.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_exception.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_util.dart';

import 'package:synchronized/synchronized.dart';
// alternative: import 'package:synchronized_lite/synchronized_lite.dart';

abstract class SuRecordRepo {
  Future<SuRecordFile> save();
  Future<List<SuRecordFile>> findAll();
  Future<SuRecordFile> findByID(String id);
  void delete(String id);
  Future<bool> existsById(String id);
  Future<int> count();
}


class SuRecordFilesRepo extends SuRecordRepo {
  final log = Logger('SuRecordFilesRepo');

  static final SuRecordFilesRepo _suRecordFilesRepo = SuRecordFilesRepo._internal();
  static Lock _repoLock = new Lock(reentrant: false);
  static List<SuRecordFile> _suRecordFiles = [];

  SuRecordFilesRepo._internal();

  factory SuRecordFilesRepo() {
    return _suRecordFilesRepo;
  }

  void initializeSuRecordFiles(List<SuRecordFile> suRecordFiles) {
    _suRecordFiles = suRecordFiles;
  }

  
  Future<SuRecordFile> _saveData(
      Uint8List data, List<String> tags, List<String> refs, SuRecordType suRecordType) async {

    if (data != null) {

      // this could be a data duplicate and we check for that
      SuRecord localSuRecord = await SuRecordUtil.createLocalSuRecord(data, tags, refs, suRecordType);
      String localSuRecordId = SuRecordUtil.idHash_fromSuRecord(localSuRecord);
      bool duplicate = _existsById(localSuRecordId);

      if (!duplicate) {
        SuRecordFile suRecordFile = await SuRecordUtil.createLocalSuRecordFileFromLocalSuRecord(data, localSuRecord);
        return Future.value(suRecordFile);

      } else {
        final message  = "_saveData was called with a data duplicate";
        log.severe(message);
        throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
      }
    } else {
      final message = "_saveData was called without data";
      log.severe(message);
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    }
  }


  Future<SuRecordFile> _updateMetadata(SuRecordFile suRecordFile) async {

    if (suRecordFile != null) {
      return SuRecordUtil.updateMetadata(suRecordFile);
    } else {
      final message = "_updateMetadata was called without a record to update";
      log.severe(message);
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    }
  }


  /// A new suRecordFile is created if param suRecordFile is null.
  /// - the tags, refs, verifications are used for the creation.
  ///
  /// If suRecordFile is not null an update (overwrite, not merge) is made.
  /// - the tags, refs, verifications are used for replacing the old data.
  ///
  /// The repo memory list and repo timestamp are
  /// updated according to the created or updated record.
  ///
  Future<SuRecordFile> save({
    Uint8List data,
    List<String> tags,
    List<String> refs,
    List<Verification> verifications,
    SuRecordType suRecordType,
    SuRecordFile suRecordFile
  }) async {
    return await _repoLock.synchronized(() async {

      if (suRecordFile != null) { // ~ update

        // TODO: doublecheck if passed in data is the same as in an existing surecord ?

        String toBeUpatedSuRecordFileId = SuRecordUtil.idHash_fromSuRecordFile(suRecordFile);
        int originalSuRecordFileIndex = _suRecordFiles.indexWhere(
                (suRecordFile) =>
                (0 == toBeUpatedSuRecordFileId.compareTo(SuRecordUtil.idHash_fromSuRecordFile(suRecordFile))));
        if (-1 == originalSuRecordFileIndex) {
          final message = "save was called for update of a not existing record";
          log.severe(message);
          throw new SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
        }

        suRecordFile.suRecord.tags = tags;
        suRecordFile.suRecord.refs = refs;
        suRecordFile.suRecord.verifications = verifications;

        // TODO: update timestamps incl repo

        SuRecordFile updatedSuRecordFile = await _updateMetadata(suRecordFile);

        // TODO: specify sort order and  handle sorting if necessary ?
        _suRecordFiles[originalSuRecordFileIndex] = updatedSuRecordFile;

        return updatedSuRecordFile;

      } else if (data != null && suRecordType != null) { // ~ create

        // TODO: update timestamps incl repo

        SuRecordFile newSuRecordFile = await _saveData(data, tags, refs, suRecordType);

        // TODO: specify sort order and handle sorting if necessary ?
        _suRecordFiles.add(newSuRecordFile);

        return newSuRecordFile;

      } else {
        final message = "save was called with params not matching update or create";
        log.severe(message);
        throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
      }
    });
  }


  Future<bool> delete(String id) async {
    return await _repoLock.synchronized(() async {

      int index = _suRecordFiles.indexWhere(
              (suRecordFile) =>
              (0 == id.compareTo(SuRecordUtil.idHash_fromSuRecordFile(suRecordFile))));
      SuRecordFile suRecordFile = _suRecordFiles[index];
      String filename = suRecordFile.surecFilename;
      File file = File(filename);
      bool deleted = true;
      try {
        file.deleteSync();
        _suRecordFiles.removeAt(index);
      } catch (e, s) {
        final message = "delete lead to an exception caused by params or other reason";
        log.severe(message);
        throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
      }
      return Future.value(deleted);
    });
  }


  List<SuRecordFile> _findAllSync()  {
    return _suRecordFiles;
  }


  Future<List<SuRecordFile>> findAll() async {
    return await _repoLock.synchronized(() async {
      return Future.value(_findAllSync());
    });
  }


  /**
   * Not protected by lock -> only internal use
   * returns record if record with id is found
   * throws otherwise
   */
  SuRecordFile _findByID(String id) {
    SuRecordFile foundSuRecordFile = null;
    if (_suRecordFiles.isNotEmpty) {
      foundSuRecordFile = _suRecordFiles.firstWhere(
              (suRecordFile) =>
              (0 == id.compareTo(SuRecordUtil.idHash_fromSuRecordFile(suRecordFile))),
          orElse: () => null);
    }
    if (null == foundSuRecordFile) {
      final message = "_findByID was called with an invalid ID";
      log.warning(message);
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    }
    return foundSuRecordFile;
  }


  Future<SuRecordFile> findByID(String id) async {
    return await _repoLock.synchronized(() async {
      return Future.value(_findByID( id));
    });
  }


  bool _existsById(String id)  {
    SuRecordFile suRecordFile = null;
    try {
      suRecordFile = _findByID(id);
    } catch (e) {
      final message = "_existsById was called with not existing ID ";
      log.info(message);
    }
    return (suRecordFile != null ? true : false);
  }


  Future<bool> existsById(String id) async {
    return await _repoLock.synchronized(() async {
      return Future.value(_existsById(id));
    });
  }


  Future<int> count() async {
    return await _repoLock.synchronized(() async {
      return Future.value(_suRecordFiles.length);
    });
  }


  Future<SuRecordFilesRepoStatistics> statistics() async {
    return await _repoLock.synchronized(() async {

      Set<String> tagSet = Set<String>();
      Set<String> refSet = Set<String>();
      for (var suRecordFile in _suRecordFiles) {
        tagSet.addAll(suRecordFile.suRecord.tags);
        refSet.addAll(suRecordFile.suRecord.refs);
      }
      return Future.value(SuRecordFilesRepoStatistics(_suRecordFiles.length, tagSet, refSet));
    });
  }
}


class SuRecordFilesRepoStatistics extends Equatable {
  final int _count;
  int get count => _count;

  final Set<String> _tags;
  Set<String> get tags => _tags;

  final Set<String> _refs;
  Set<String> get refs => _refs;

  const SuRecordFilesRepoStatistics(this._count, this._tags, this._refs);

  @override
  List<Object> get props => [count, tags];
}
