/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'package:archive/archive.dart';
import 'package:archive/archive_io.dart';

import 'package:convert/convert.dart';
import 'package:flutter/foundation.dart';
import 'package:logging/logging.dart';

import 'package:path/path.dart' as p;

import 'package:pointycastle/pointycastle.dart';
import 'package:surec_app/surecord/surecord_config.dart';

import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_exception.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_webserviceutil.dart'; // see importing https://pub.dev/packages/pointycastle



class SuRecordUtil {
  static final log = Logger('SuRecordBlocRepo');

  static final SORT_ASCENDING = true;

  /// this method prepends the path from LocalSuRecordFilesConfig.pathToSuRecordFiles
  /// to the filename and returns the Future<File>
  static Future<File> fileForName(String filename) async {
    log.info("fileForName( $filename )");

    String path = await LocalSuRecordFilesConfig.pathToSuRecordFiles;
    String separator = p.separator;
    String pathAndFilename = '$path$separator$filename';
    File file = new File(pathAndFilename);
    return Future.value(file);
  }


  static Future<List<Fingerprint>> createFingerprintList(Uint8List data, List<Algo> algos) async {
    List<Fingerprint> resultL = [];
    for (Algo algo in algos) {
      Digest digest = Digest(algoToSring(algo));
      Uint8List value = digest.process(data);
      resultL.add(Fingerprint(algo, hex.encode(value)));
    }
    return resultL;
  }


  /// a confirmation is calculated for a list of fingerprints
  /// with the same uri and the same data but with different
  /// fingerprint algorithms
  static Confirmation createConfirmation(List<Fingerprint> fingerprints, Algo algo, int timestamp) {
    String concatenated = "";
    for (Fingerprint fingerprint in fingerprints) {
      concatenated += fingerprint.value;
    }
    Digest digest = Digest(algoToSring(algo));
    Uint8List data = utf8.encode(concatenated);
    String value = hex.encode(digest.process(data));
    return Confirmation(value, algo, timestamp);
  }


  static createLocalVerification(Uint8List data) async {

    SuRecordVerificationConfig verifierConfig = LocalSuRecordVerificationConfig();

    String uri = verifierConfig.verifierURI;
    List<Algo> algos = verifierConfig.verifcationAlgos;

    int timestamp = DateTime.now().millisecondsSinceEpoch;
    // 32 bit utc system time - option: use this for sort by time with
    List<Fingerprint> fingerprints = await createFingerprintList(data, algos);
    Confirmation confirmation = createConfirmation(fingerprints, verifierConfig.confirmationAlgo, timestamp);
    Verification verification = Verification(uri, timestamp, fingerprints, confirmation);
    return verification;
  }


  static Future<SuRecord> createLocalSuRecord(
      Uint8List data, List<String> tags, List<String> refs, SuRecordType suRecordType) async {
    SuRecordVerificationConfig localSuRecordVerificationConfig = LocalSuRecordVerificationConfig();

    Verification localVerification = await createLocalVerification(data);

    String idAlgo;
    List<Fingerprint> fingerprints = localVerification.fingerprints;
    for (Fingerprint fingerprint in fingerprints) {
      if (fingerprint.algo == localSuRecordVerificationConfig.idAlgo) {
        idAlgo = fingerprint.value;
      }
    }
    String dataFilenamePrefix = idAlgo;
    String dataFilenameSuffix = suRecordTypeToString(suRecordType);
    String dataFilename = '$dataFilenamePrefix.$dataFilenameSuffix';

    var modificationTime = localVerification.timestamp;
    SuRecord localSuRecord = SuRecord([localVerification], tags, refs, modificationTime, suRecordType, dataFilename);
    return localSuRecord;
  }


  /// Creates a SuRecord file in the local filesytem
  ///
  /// @param data   Content of the file that will be verifiable and that can not be changed
  /// @params tags  Changeable metadata for the verfiable file content.
  /// @params refs  Changeable references to other SuRecord files.
  /// @params filetype of the content file in the SurRecord file.
  /// @params filename of the content file in the SurRecord file. Should and will be generated if left blank or null.
  ///
  static Future<SuRecordFile> createLocalSuRecordFile(
      Uint8List data, List<String> tags, List<String> refs, SuRecordType suRecordType) async {
    SuRecord localSuRecord = await createLocalSuRecord(data, tags, refs, suRecordType);
    SuRecordFile localSuRecordFile = await writeSuRecordFile(localSuRecord, data);

    return Future.value(localSuRecordFile);
  }


  static Future<SuRecordFile> createLocalSuRecordFileFromLocalSuRecord(Uint8List data, SuRecord localSuRecord) async {

    try {
      Verification verification = await SuRecordWebserviceUtil.postFingerprintsForConfirmation(localSuRecord);
      localSuRecord.verifications.add(verification);
    } catch (e,s) {
      log.info("could nod confirm");
    }

    SuRecordFile suRecordFile = await writeSuRecordFile(localSuRecord, data);
    return Future.value(suRecordFile);
  }


  static Future<SuRecordFile> writeSuRecordFile(SuRecord localSuRecord, Uint8List data) async {

    String dataFilename = localSuRecord.dataFilename;
    ArchiveFile dataArchiveFile = ArchiveFile("$dataFilename", data.length, data);

    Uint8List metadata = utf8.encode(jsonEncode(localSuRecord)) as Uint8List;
    ArchiveFile metadataArchiveFile = ArchiveFile(LocalSuRecordFilesConfig().metadataFilename, metadata.length, metadata);

    String archiveFilenamePrefix = localSuRecord.modificationTime.toRadixString(16).toUpperCase();
    String archiveFilenameInfix = suRecordTypeToString(localSuRecord.type);
    String archiveFilenameSuffix = LocalSuRecordFilesConfig().suRecordPostfix;

    String archiveFilename = '$archiveFilenamePrefix.$archiveFilenameInfix.$archiveFilenameSuffix';

    Archive archive = Archive();
    archive.addFile(dataArchiveFile);
    archive.addFile(metadataArchiveFile);

    final zipData = ZipEncoder().encode(archive);

    try {
      String writePath = await LocalSuRecordFilesConfig.pathToSuRecordFiles;
      String separator = p.separator;
      String suRecordFileName = '$writePath$separator$archiveFilename';

      File f = new File('$writePath$separator$archiveFilename');
      await f.writeAsBytes(zipData, flush: true);

      SuRecordFile suRecordFile = SuRecordFile(localSuRecord, suRecordFileName);
      return Future.value(suRecordFile);
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "writeSuRecordFile write to filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  static String idHash_fromSuRecord(SuRecord suRecord) {
    String idHash = suRecord.dataFilename.substring(
        0, suRecord.dataFilename.length - suRecordTypeToString(suRecord.type).length + 1);
    return idHash;
  }


  static String idHash_fromSuRecordFile(SuRecordFile suRecordFile) {
    String idHash = idHash_fromSuRecord(suRecordFile.suRecord);
    return idHash;
  }


  static Future<List<String>> scanForSuRecordFiles() async {
    log.info("scanForSuRecordFiles");
    List<String> resultL = [];
    try {
      String scanPath = await LocalSuRecordFilesConfig.pathToSuRecordFiles;
      log.info("scanForSuRecordFiles scanpath $scanPath");

      final context = p.Context();
      Directory dir = Directory(scanPath);

      Stream<FileSystemEntity> fileSystemEntityStream = dir.list(recursive: false, followLinks: false);
      await for (FileSystemEntity entity in fileSystemEntityStream) {
        if (context.extension(entity.path) == "." + LocalSuRecordFilesConfig().suRecordPostfix) {
          resultL.add(entity.path);
          log.info("scanForSuRecordFiles found:" + entity.path);
        } else {
          log.info("scanForSuRecordFiles miss:" + entity.path);
        }
      }
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "scanForSuRecordFiles reading filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
    resultL.sort((a, b) => SORT_ASCENDING ? a.compareTo(b) : b.compareTo(a));
    return Future.value(resultL);
  }


  static Future<List<SuRecordFile>> listSuRecordFiles() async {
    log.info("listSuRecordFiles");

    List<String> suRecordFilenameList = await scanForSuRecordFiles();

    List<SuRecordFile> suRecordFiles = [];

    for (String suRecordFilename in suRecordFilenameList) {
      try {
        File f = new File(suRecordFilename);
        List<int> bytes = await f.readAsBytes();
        Archive archive = new ZipDecoder().decodeBytes(bytes);

        SuRecord latestSuRecord;
        var latestSuRecordModificationTimestamp = 0;

        for (ArchiveFile inArchiveFile in archive) {
          String filename = inArchiveFile.name;
          List<SuRecord> suRecordsInArchive = [];
          if (filename == LocalSuRecordFilesConfig().metadataFilename) {
            List<int> metadata = inArchiveFile.content;
            String metadataJS = utf8.decode(metadata);
            var metatdataJson = json.decode(metadataJS);
            SuRecord suRecord = SuRecord.fromJson(metatdataJson);
            if(suRecord.modificationTime > latestSuRecordModificationTimestamp) {
              latestSuRecordModificationTimestamp = suRecord.modificationTime;
              latestSuRecord = suRecord;
            }
          }
        }
        SuRecordFile suRecordFile = SuRecordFile(latestSuRecord, suRecordFilename);
        suRecordFiles.add(suRecordFile);
      } catch (exception, stacktrace) {
        final exceptionStr = exception.toString();
        final stacktraceStr = stacktrace.toString();
        final message = "writeSuRecordFile write to filesystem failed";
        log.severe("$message\n$exceptionStr\n$stacktraceStr");
        throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
      }
    }
    suRecordFiles.sort((a, b) => SORT_ASCENDING
        ? localConfirmationTimestamp(a.suRecord).compareTo(localConfirmationTimestamp(b.suRecord))
        : localConfirmationTimestamp(b.suRecord).compareTo(localConfirmationTimestamp(a.suRecord)) );
    return Future.value(suRecordFiles);
  }


  static Future<SuRecordFile> updateMetadata(SuRecordFile suRecordFile) {
    log.info("updateMetadata");

    try {
      int modificationTimestamp = DateTime.now().millisecondsSinceEpoch;
      suRecordFile.suRecord.modificationTime = modificationTimestamp;


      Uint8List metadata = utf8.encode(jsonEncode(suRecordFile.suRecord)) as Uint8List;
      Uint8List data;

      final oldBytes = File(suRecordFile.surecFilename).readAsBytesSync();

      ArchiveFile metadataArchiveFile = ArchiveFile.noCompress(
          LocalSuRecordFilesConfig().metadataFilename,
          metadata.length, metadata);

      final archive = ZipDecoder().decodeBytes(oldBytes);

      archive.addFile(metadataArchiveFile);

      final newZipData = ZipEncoder().encode(archive);
      final newFile = File(suRecordFile.surecFilename);
      newFile.writeAsBytesSync(newZipData, flush: true);

      return Future.value(suRecordFile);

    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "updateMetadata in zip on filesystem failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  static Uint8List dataFromSuRecordFileAsMem(SuRecordFile suRecordFile) {
    log.info("dataFromSuRecordFileAsMem()");

    try {
      Uint8List data;
      final fileBytes = File(suRecordFile.surecFilename).readAsBytesSync();

      final archive = ZipDecoder().decodeBytes(fileBytes);
      for (final file in archive) {
        final filename = file.name;
        if (file.isFile && (file.name.compareTo(suRecordFile.suRecord.dataFilename) == 0 )) {
          data = file.content as Uint8List;
          break;
        }
      }
      return data;

    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "dataFromSuRecordFileAsMem read failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  static void dataFromSuRecordFileAsFile(SuRecordFile suRecordFile, String pathWithName) {
    log.info("dataFromSuRecordFileAsFile()");
    try {
      var targetFile = File(pathWithName);
      var data = dataFromSuRecordFileAsMem(suRecordFile);
      targetFile.writeAsBytesSync(data, flush: true);

    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "dataFromSuRecordFileAsFile extract and write failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  static Future<Uint8List> dataFromSuRecordFileAsFuture(SuRecordFile suRecordFile) async {
    log.info("dataFromSuRecordFileAsFuture");

    try {
      return compute(dataFromSuRecordFileAsMem,suRecordFile);
    } catch (exception, stacktrace) {
      final exceptionStr = exception.toString();
      final stacktraceStr = stacktrace.toString();
      final message = "dataFromSuRecordFileAsMem read failed";
      log.severe("$message\n$exceptionStr\n$stacktraceStr");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  static int lastModification(SuRecord suRecord) {
    return suRecord.modificationTime;
  }


  static int localConfirmationTimestamp(SuRecord suRecord) {
    SuRecordVerificationConfig localSuRecordVerifierConfig = LocalSuRecordVerificationConfig();

    var first;
    for(var verification in suRecord.verifications) {
      if (localSuRecordVerifierConfig.verifierURI.compareTo(verification.uri) == 0) {
        first = verification.timestamp;
      }
    }
    if(first == null) {
      final message = "localConfirmationTimestamp surecord has no local confirmation timestamp";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    } else {
      return first;
    }
  }


  static int lastRemoteConfirmationTimestamp(SuRecord suRecord) {
    SuRecordVerificationConfig localSuRecordVerifierConfig = LocalSuRecordVerificationConfig();

    var last;
    for(var verification in suRecord.verifications) {
      if (localSuRecordVerifierConfig.verifierURI.compareTo(verification.uri) != 0) {
        if(verification.timestamp > last) {
          last = verification.timestamp;
        }
      }
    }
    if(last == null) {
      final message = "lastRemoteConfirmationTimestamp surecord  has no last remote confirmation timestamp";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    } else {
      return last;
    }
  }


  static int firstRemoteConfirmationTimestamp(SuRecord suRecord) {
    SuRecordVerificationConfig localSuRecordVerifierConfig = LocalSuRecordVerificationConfig();

    var first;
    for(var verification in suRecord.verifications) {
      if (localSuRecordVerifierConfig.verifierURI.compareTo(verification.uri) != 0) {
        if (first==-1) {
          first = verification.timestamp;
        } else if(verification.timestamp > first) {
          first = verification.timestamp;
        }
      }
    }
    if(first == null) {
      final message = "firstRemoteConfirmationTimestamp surecord has no first remote confirmation timestamp";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    } else {
      return first;
    }
  }


  static String generateCacheEntryStrippedname(SuRecord suRecord) {
   return localConfirmationTimestamp(suRecord).toRadixString(16).toUpperCase();
  }
}

