/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import 'package:path/path.dart' as p;
import 'package:logging/logging.dart';
import 'package:surec_app/cache/fs_audio_cache.dart';
import 'package:surec_app/cache/fs_image_and_video_cache.dart';
import 'package:surec_app/cache/fs_note_cache.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_repo.dart';
import 'package:surec_app/surecord/surecord_util.dart';
import 'package:video_thumbnail/video_thumbnail.dart';

abstract class BlocSuRecordDataFacade {

  BlocSuRecordDataFacade();

  dataForBlocSuRecord(BlocSuRecord blocSuRecord, bool preview);

  Future<String> _textFromPathWithFilename(String pathWithFilename) async {
    File textFile = new File(pathWithFilename);
    Future<String> textString = textFile.readAsString();
    return Future.value(textString);
  }


  Future<String> _textFromPathWithFilenameSync(String pathWithFilename) async {
    File textFile = new File(pathWithFilename);
    String textString = textFile.readAsStringSync();
    return textString;
  }
}

class AudioBlocSuRecordDataFacade extends BlocSuRecordDataFacade{
  static final log = Logger('VideoBlocSuRecordDataFacade');

  final validSuRecordTypes = [SuRecordType.mp3, SuRecordType.aac];
  final int _thumbnailSize = 192;
  final _tempSuffix = ".temp";

  final FSAudioCache _fsAudioCache;
  final  _tempDirPath;

  AudioBlocSuRecordDataFacade(FSAudioCache fsAudioCache, String tempDirPath) :
        this._fsAudioCache = fsAudioCache,
        this._tempDirPath = tempDirPath {
  }


  Future<String> dataForBlocSuRecord(BlocSuRecord blocSuRecord, preview) async {
    
    var suRecordFile = await SuRecordFilesRepo().findByID(blocSuRecord.id);

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsAudioCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if ( ! preview) {
          var pathToKnownItem = _fsAudioCache.getPathToKnownItemForStrippedName(cacheEntryStrippedname);
          if(pathToKnownItem != null) {
            return Future.value(pathToKnownItem);
          }
        }
      }

      if ( ! preview) {
        var tempfilenname  = DateTime.now().millisecondsSinceEpoch.toString() + _tempSuffix;
        var tempPathWithFilename = p.join(_tempDirPath.toString(), tempfilenname);

        SuRecordUtil.dataFromSuRecordFileAsFile(suRecordFile, tempPathWithFilename);
        return Future.value(tempPathWithFilename);
      }
    }
  }
}


class NoteBlocSuRecordDataFacade extends BlocSuRecordDataFacade{

  final validSuRecordTypes = [SuRecordType.txt, SuRecordType.md];
  final int _textsnippetSize = 400;

  final FSNoteCache _fsNoteCache;
  var _tempDir;
  final  _tempDirPath;

  NoteBlocSuRecordDataFacade(FSNoteCache fsNoteCache, tempDirPath) :
        this._fsNoteCache = fsNoteCache,
        this._tempDirPath = tempDirPath {
  }


  Future<String> dataForBlocSuRecord(BlocSuRecord blocSuRecord, preview) async {

    var suRecordFile = await SuRecordFilesRepo().findByID(blocSuRecord.id);

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsNoteCache != null) {
        var cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownTextsnippet = _fsNoteCache.getPathToKnownTextsnippetForStrippedName(cacheEntryStrippedname);
          if(pathToKnownTextsnippet != null) {
            return _textFromPathWithFilename(pathToKnownTextsnippet);
          } else {
            var pathToKnownText = _fsNoteCache.getPathToKnownTextForStrippedName(cacheEntryStrippedname);
            if (pathToKnownText != null) {
              return _textFromPathWithFilename(pathToKnownTextsnippet);
            }
          }
        }
      }
      Uint8List textBytes = SuRecordUtil.dataFromSuRecordFileAsMem(suRecordFile);
      var text = Utf8Decoder().convert(textBytes);

      if(preview) {
        var textsnippet = text.substring(0, min(_textsnippetSize, text.length));
        return Future.value(textsnippet);
      }
      return Future.value(text);
    }
  }
}


class VideoBlocSuRecordDataFacade extends BlocSuRecordDataFacade{
  static final log = Logger('VideoBlocSuRecordDataFacade');

  final validSuRecordTypes = [SuRecordType.mp4];
  final String _tempSuffix=".temp";
  final int _thumbnailSize = 192;

  final FSImageAndVideoCache _fsImageAndVideoCache;
  final  _tempDirPath;

  VideoBlocSuRecordDataFacade(FSImageAndVideoCache fsImageAndVideoCache, tempDirPath) :
        this._fsImageAndVideoCache = fsImageAndVideoCache,
        this._tempDirPath = tempDirPath {
  }


  Future<String> dataForBlocSuRecord(BlocSuRecord blocSuRecord, preview) async {

    var suRecordFile = await SuRecordFilesRepo().findByID(blocSuRecord.id);


    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsImageAndVideoCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownThumbnail = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownThumbnail != null) {
            return pathToKnownThumbnail;
          }
        } 
      }

      var tempfilenname  = DateTime.now().millisecondsSinceEpoch.toString() + _tempSuffix;
      var tempPathWithFilename = p.join(_tempDirPath, tempfilenname);

      SuRecordUtil.dataFromSuRecordFileAsFile(suRecordFile, tempPathWithFilename);      

      if (preview) {
        try {
            Uint8List encodedVideoTumbnail;
            encodedVideoTumbnail = await VideoThumbnail.thumbnailData(
              video: tempPathWithFilename,
              imageFormat: ImageFormat.JPEG,
              maxWidth: _thumbnailSize,
              maxHeight: _thumbnailSize,
              quality: 60,
            );

            var file = File(tempPathWithFilename);
            file.writeAsBytesSync(encodedVideoTumbnail, flush: true);

            return tempPathWithFilename;
          
        } catch (e,s) {
          log.severe("wtf");
        }
      }
      else {
        return tempPathWithFilename;
      }

    }
  }
}


class ImageBlocSuRecordDataFacade extends BlocSuRecordDataFacade{

  final validSuRecordTypes = [SuRecordType.png, SuRecordType.jpg];   
  final String _tempSuffix=".temp";
  final int _thumbnailSize = 192;

  final FSImageAndVideoCache _fsImageAndVideoCache;
  final  _tempDirPath;

  ImageBlocSuRecordDataFacade(FSImageAndVideoCache fsImageAndVideoCache, tempDirPath) :
        this._fsImageAndVideoCache = fsImageAndVideoCache,
        this._tempDirPath = tempDirPath {
  }


  Future<String> dataForBlocSuRecord(BlocSuRecord blocSuRecord, preview) async {

    var suRecordFile = await SuRecordFilesRepo().findByID(blocSuRecord.id);

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsImageAndVideoCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownThumbnail = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownThumbnail != null) {
            return pathToKnownThumbnail;
          }
        } else {
          var pathToKnownImage = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownImage != null) {
            return pathToKnownImage;
          }
        }
      }

      var tempfilenname  = DateTime.now().millisecondsSinceEpoch.toString() + _tempSuffix;
      var tempPathWithFilename = p.join(_tempDirPath, tempfilenname);
      SuRecordUtil.dataFromSuRecordFileAsFile(suRecordFile, tempPathWithFilename);

      return tempPathWithFilename;

    }
  }
}
