/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:intl/intl.dart';
import 'dart:convert';
import 'dart:math';
import 'dart:typed_data';
import 'dart:io';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as p;
import 'package:flutter/services.dart';
import 'package:pdf/pdf.dart';
import 'package:pdf/widgets.dart' as pdfWidgets;
import 'package:image/image.dart' as imagePkg;
import 'package:surec_app/cache/fs_audio_cache.dart';
import 'package:surec_app/cache/fs_image_and_video_cache.dart';
import 'package:surec_app/cache/fs_note_cache.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_util.dart';
import 'package:video_thumbnail/video_thumbnail.dart';
import 'package:path_provider/path_provider.dart';


abstract class SuRecordDataFacade {

  SuRecordDataFacade();

  dataForSuRecordFile(SuRecordFile SuRecordFile, bool preview);

  Future<String> _textFromPathWithFilename(String pathWithFilename) async {
    File textFile = new File(pathWithFilename);
    Future<String> textString = textFile.readAsString();
    return Future.value(textString);
  }


  Future<String> _textFromPathWithFilenameSync(String pathWithFilename) async {
    File textFile = new File(pathWithFilename);
    String textString = textFile.readAsStringSync();
    return textString;
  }

  Future<pdfWidgets.ImageImage> _imageFromPathWithFilename(String pathWithFilename) async {
    File encodedImageFile = new File(pathWithFilename);
    Future<Uint8List> encodedImageBytes = encodedImageFile.readAsBytes();
    imagePkg.Image imagePkgImage = imagePkg.decodeImage(await encodedImageBytes);
    pdfWidgets.ImageImage pdfWidgetsImageImage = pdfWidgets.ImageImage(imagePkgImage);
    return Future.value(pdfWidgetsImageImage);
  }


  pdfWidgets.ImageImage _imageFromPathWithFilenameSync(String pathWithFilename){
    File encodedImageFile = new File(pathWithFilename);
    Uint8List encodedImageBytes = encodedImageFile.readAsBytesSync();
    imagePkg.Image imagePkgImage = imagePkg.decodeImage(encodedImageBytes);
    pdfWidgets.ImageImage pdfWidgetsImageImage = pdfWidgets.ImageImage(imagePkgImage);
    return pdfWidgetsImageImage;
  }
}

class AudioSuRecordDataFacade extends SuRecordDataFacade{
  static final log = Logger('VideoSuRecordDataFacade');

  final validSuRecordTypes = [SuRecordType.mp3, SuRecordType.aac];
  final int _thumbnailSize = 192;
  final _tempSuffix = ".temp";

  final FSAudioCache _fsAudioCache;
  final  _tempDirPath;

  AudioSuRecordDataFacade(FSAudioCache fsAudioCache, String tempDirPath) :
        this._fsAudioCache = fsAudioCache,
        this._tempDirPath = tempDirPath {
  }

  // not caching, only audio full  - not tested at all
  Future<String> dataForSuRecordFile(SuRecordFile suRecordFile, preview) async {

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsAudioCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if ( ! preview) {
          var pathToKnownItem = _fsAudioCache.getPathToKnownItemForStrippedName(cacheEntryStrippedname);
          if(pathToKnownItem != null) {
            return Future.value(pathToKnownItem);
          }
        }
      }

      if ( ! preview) {
        var tempfilenname  = DateTime.now().millisecondsSinceEpoch.toString() + _tempSuffix;
        var tempPathWithFilename = p.join(_tempDirPath.toString(), tempfilenname);

        SuRecordUtil.dataFromSuRecordFileAsFile(suRecordFile, tempPathWithFilename);
        return Future.value(tempPathWithFilename);
      }
    }
  }
}


class NoteSuRecordDataFacade extends SuRecordDataFacade{

  final validSuRecordTypes = [SuRecordType.txt, SuRecordType.md];
  final int _textsnippetSize = 400;

  final FSNoteCache _fsNoteCache;
  var _tempDir;
  final  _tempDirPath;

  NoteSuRecordDataFacade(FSNoteCache fsNoteCache, tempDirPath) :
        this._fsNoteCache = fsNoteCache,
        this._tempDirPath = tempDirPath {
  }

  // not caching, only images
  Future<String> dataForSuRecordFile(SuRecordFile suRecordFile, preview) async {

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsNoteCache != null) {
        var cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownTextsnippet = _fsNoteCache.getPathToKnownTextsnippetForStrippedName(cacheEntryStrippedname);
          if(pathToKnownTextsnippet != null) {
            return _textFromPathWithFilename(pathToKnownTextsnippet);
          } else {
            var pathToKnownText = _fsNoteCache.getPathToKnownTextForStrippedName(cacheEntryStrippedname);
            if (pathToKnownText != null) {
              return _textFromPathWithFilename(pathToKnownTextsnippet);
            }
          }
        }
      }
      Uint8List textBytes = SuRecordUtil.dataFromSuRecordFileAsMem(suRecordFile);
      var text = Utf8Decoder().convert(textBytes);

      if(preview) {
        var textsnippet = text.substring(0, min(_textsnippetSize, text.length));
        return Future.value(textsnippet);
      }
      return Future.value(text);
    }
  }
}


class VideoSuRecordDataFacade extends SuRecordDataFacade{
  static final log = Logger('VideoSuRecordDataFacade');

  final validSuRecordTypes = [SuRecordType.mp4];
  final String _tempSuffix=".temp";
  final int _thumbnailSize = 192;

  final FSImageAndVideoCache _fsImageAndVideoCache;
  final  _tempDirPath;

  VideoSuRecordDataFacade(FSImageAndVideoCache fsImageAndVideoCache, tempDirPath) :
        this._fsImageAndVideoCache = fsImageAndVideoCache,
        this._tempDirPath = tempDirPath {
  }

  // not caching, only images
  Future<pdfWidgets.ImageImage> dataForSuRecordFile(SuRecordFile suRecordFile, preview) async {


    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsImageAndVideoCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownThumbnail = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownThumbnail != null) {
            return _imageFromPathWithFilename(pathToKnownThumbnail);
          }
        }
      }

      var tempfilenname  = DateTime.now().millisecondsSinceEpoch.toString() + _tempSuffix;
      var tempPathWithFilename = p.join(_tempDirPath, tempfilenname);

      SuRecordUtil.dataFromSuRecordFileAsFile(suRecordFile, tempPathWithFilename);

      Uint8List encodedVideoTumbnail;
      try {
        if (preview) {
          encodedVideoTumbnail = await VideoThumbnail.thumbnailData(
            video: tempPathWithFilename,
            imageFormat: ImageFormat.JPEG,
            maxWidth: _thumbnailSize,
            maxHeight: _thumbnailSize,
            quality: 60,
          );
        } else {
          encodedVideoTumbnail = await VideoThumbnail.thumbnailData(
            video: tempPathWithFilename,
            imageFormat: ImageFormat.JPEG,
            quality: 60,
          );
        }
        imagePkg.Image imagePkgImage = imagePkg.decodeImage(encodedVideoTumbnail);
        pdfWidgets.ImageImage pdfWidgetsImageImage = pdfWidgets.ImageImage(imagePkgImage);
        return pdfWidgetsImageImage;
      } catch (e,s) {
        log.severe("wtf");
      }
    }
  }
}


class ImageSuRecordDataFacade extends SuRecordDataFacade{

  final validSuRecordTypes = [SuRecordType.png, SuRecordType.jpg];
  final String _tempSuffix=".temp";
  final int _thumbnailSize = 192;

  final FSImageAndVideoCache _fsImageAndVideoCache;
  final  _tempDirPath;

  ImageSuRecordDataFacade(FSImageAndVideoCache fsImageAndVideoCache, tempDirPath) :
        this._fsImageAndVideoCache = fsImageAndVideoCache,
        this._tempDirPath = tempDirPath {
  }


  // not caching, only images
  Future<pdfWidgets.ImageImage> dataForSuRecordFile(SuRecordFile suRecordFile, preview) async {

    if( suRecordFile.suRecord != null &&
        validSuRecordTypes.contains(suRecordFile.suRecord.type) ) {

      if(_fsImageAndVideoCache != null) {
        String cacheEntryStrippedname = SuRecordUtil.generateCacheEntryStrippedname(suRecordFile.suRecord);
        if (preview) {
          var pathToKnownThumbnail = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownThumbnail != null) {
            return _imageFromPathWithFilename(pathToKnownThumbnail);
          }
        } else {
          var pathToKnownImage = _fsImageAndVideoCache.getPathToKnownThumbnailForStrippedName(cacheEntryStrippedname);
          if(pathToKnownImage != null) {
            return _imageFromPathWithFilename(pathToKnownImage);
          }
        }
      }

      Uint8List mem = SuRecordUtil.dataFromSuRecordFileAsMem(suRecordFile);
      var imagePkgImage = imagePkg.decodeImage(mem);

      if(preview) {
        imagePkg.Image imagePkgImageThmb = imagePkg.copyResizeCropSquare(imagePkgImage, _thumbnailSize);
        pdfWidgets.ImageImage pdfWidgetsImageImageThmb = pdfWidgets.ImageImage(imagePkgImageThmb);
        return Future.value(pdfWidgetsImageImageThmb);
      } else {
        pdfWidgets.ImageImage pdfWidgetsImageImage = pdfWidgets.ImageImage(imagePkgImage);
        return Future.value(pdfWidgetsImageImage);
      }
    }
  }
}


class PdfExporter {
  static const FontSizeXXS = 5.0;
  static const FontSizeXS = 6.0;
  static const FontSizeS = 8.0;
  static const FontSizeM = 10.0;
  static const FontSizeL = 12.0;
  static const FontSizeXL = 14.0;

  static const format_insets = 0.2 * PdfPageFormat.cm;
  static const format_insetsSmall = 0.1 * PdfPageFormat.cm;
  static const format_insetsXtraSmall = 0.06 * PdfPageFormat.cm;

  static final format_pageformat = PdfPageFormat.a4;

  static const edgeInsets = pdfWidgets.EdgeInsets.only(
      left: format_insets,
      top: format_insets,
      bottom: format_insets,
      right: format_insets);
  static const edgeInsetsSmall = pdfWidgets.EdgeInsets.only(
      left: format_insetsSmall,
      top: format_insetsSmall,
      bottom: format_insetsSmall,
      right: format_insetsSmall);
  static const edgeInsetsXtraSmall = pdfWidgets.EdgeInsets.only(
      left: format_insetsXtraSmall,
      top: format_insetsXtraSmall,
      bottom: format_insetsXtraSmall,
      right: format_insetsXtraSmall);

  static final webCheckUrlPrefix = "https://surec.app?confirmationValue=";


  pdfWidgets.Document _pdf;
  final List<SuRecordFile> _suRecordFiles;
  final String _targetPathWithFilename;

  PdfExporter(this._suRecordFiles, this._targetPathWithFilename) {
    _pdf = pdfWidgets.Document(title: 'Export', author: 'surec app');
  }


  pdfWidgets.ImageImage dummy_pwImageImage;
  SuRecord dummyLocalSuRecord;
  SuRecordFile dummyLocalSuRecordFile;


  Future<bool> dummyInit() async {
    Uint8List jpegEncodedUInt8List = (await rootBundle.load('devassets/3840x2180div2.jpg')).buffer.asUint8List();
    dummyLocalSuRecord = await SuRecordUtil.createLocalSuRecord(
        jpegEncodedUInt8List, ["tag1"], ["ref1"], SuRecordType.jpg);
    dummyLocalSuRecordFile = SuRecordFile(dummyLocalSuRecord, "dummyfilename");
    imagePkg.Image img_Image = imagePkg.decodeImage(jpegEncodedUInt8List);
    dummy_pwImageImage = pdfWidgets.ImageImage(img_Image); //, PdfImageOrientation.leftTop, 72.00);
    return Future.value(true);
  }


  Future<String> writePdf(List<SuRecordFile> suRecordFiles, pathWithFilename) async {
    Directory tempDir = await getTemporaryDirectory();
    String tempPath = tempDir.path;

    ImageSuRecordDataFacade imageSuRecordDataFacade = ImageSuRecordDataFacade(null, tempPath); // todo : use cache
    VideoSuRecordDataFacade videoSuRecordDataFacade = VideoSuRecordDataFacade(null, tempPath); // todo : use cache
    NoteSuRecordDataFacade noteSuRecordDataFacade = NoteSuRecordDataFacade(null, tempPath); // todo : use cache
    AudioSuRecordDataFacade audioSuRecordDataFacade = AudioSuRecordDataFacade(null, tempPath); // todo : use cache

    List<pdfWidgets.Widget> pdfWidgetsWidgets = [
      pdfWidgets.Header(level: 0, text: 'SuRecord export'),
    ];

    for (var suRecordFile in suRecordFiles) {
      if (suRecordFile.suRecord != null) {
        if (imageSuRecordDataFacade.validSuRecordTypes.contains(suRecordFile.suRecord.type)) {
          var img = await imageSuRecordDataFacade.dataForSuRecordFile(suRecordFile, false);
          pdfWidgetsWidgets.add(SuRecordImageWidget(suRecordFile.suRecord, img));
        }
        else if (videoSuRecordDataFacade.validSuRecordTypes.contains(suRecordFile.suRecord.type)) {
          var img = await videoSuRecordDataFacade.dataForSuRecordFile(suRecordFile, false);
          pdfWidgetsWidgets.add(SuRecordImageWidget(suRecordFile.suRecord, img));
        }
        else if (noteSuRecordDataFacade.validSuRecordTypes.contains(suRecordFile.suRecord.type)) {
          var text = await noteSuRecordDataFacade.dataForSuRecordFile(suRecordFile, false);
          pdfWidgetsWidgets.add(SuRecordNoteWidget(suRecordFile.suRecord, text));
        }
        else if (audioSuRecordDataFacade.validSuRecordTypes.contains(suRecordFile.suRecord.type)) {
          var icon = pdfWidgets.Icon(pdfWidgets.IconData(0xe1b8), size: 40.0);
          pdfWidgetsWidgets.add(SuRecordIconWidget(suRecordFile.suRecord, icon));
        }
      }
    }

    final pageTheme = await _pageTheme(format_pageformat);

    _pdf.addPage(pdfWidgets.MultiPage(
      pageTheme: pageTheme,
      // margin: pdfWidgets.EdgeInsets.all(format_pageinsets),
      build: (pdfWidgets.Context context) {
        return pdfWidgetsWidgets;
      },
    ));

    File file = File(pathWithFilename);
    file.writeAsBytesSync(await _pdf.save(), flush: true);
  }

  // ----------------

  // based one resume.dart

  Future<pdfWidgets.PageTheme> _pageTheme(PdfPageFormat format) async {
    format = format.applyMargin(
        left: 0.1 * PdfPageFormat.cm,
        top: 0.1 * PdfPageFormat.cm,
        right: 0.1 * PdfPageFormat.cm,
        bottom: 0.1 * PdfPageFormat.cm);

    var materialIconsRegularByteData = await rootBundle.load('assets/MaterialIcons-Regular.ttf');

    return pdfWidgets.PageTheme(
      pageFormat: format,
      theme: pdfWidgets.ThemeData.withFont(
        icons: pdfWidgets.Font.ttf(materialIconsRegularByteData),
      ),
    );
  }

  // ----------------


  pdfWidgets.Widget SuRecordImageWidget(SuRecord suRecord, pwImageImage_img_Image) {
    return
      pdfWidgets.Container(
          padding: edgeInsetsSmall,
          color: PdfColors.grey50,
          child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              suRecordMetaDataWidget(suRecord),
              imgDataWidget(pwImageImage_img_Image),
              ...VerificationsWidget(suRecord),
            ],
          )
      );
  }

  pdfWidgets.Widget SuRecordNoteWidget(SuRecord suRecord, String text) {
    return
      pdfWidgets.Container(
          padding: edgeInsetsSmall,
          color: PdfColors.grey50,
          child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              suRecordMetaDataWidget(suRecord),
              noteDataWidget(text),
              ...VerificationsWidget(suRecord),
            ],
          )
      );
  }


  pdfWidgets.Widget SuRecordIconWidget(SuRecord suRecord, pdfWidgets.Icon icon) {
    return
      pdfWidgets.Container(
          padding: edgeInsetsSmall,
          color: PdfColors.grey50,
          child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              suRecordMetaDataWidget(suRecord),
              iconDataWidget(icon),
              ...VerificationsWidget(suRecord),
            ],
          )
      );
  }


  static pdfWidgets.Widget imgDataWidget(pwImageImage_img_Image) {
    return
      pdfWidgets.Container(
        padding: edgeInsetsSmall,
          height: 3.0 * PdfPageFormat.cm,
          width: 3.0 * PdfPageFormat.cm,
          alignment: pdfWidgets.Alignment.center,
          child: pdfWidgets.Image(
            pwImageImage_img_Image,
            height: 3.0 * PdfPageFormat.cm,
            width: 3.0 * PdfPageFormat.cm,
            alignment: pdfWidgets.Alignment.center,
          )
        );
  }


  static pdfWidgets.Widget noteDataWidget(String note) {
    return
      pdfWidgets.Container(
          padding: edgeInsetsSmall,
          height: 4.0 * PdfPageFormat.cm,
          width: 16.0 * PdfPageFormat.cm,
          child:  blackText(note, FontSizeM )
      );
  }


  static pdfWidgets.Widget iconDataWidget(pdfWidgets.Icon icon) {
    return
      pdfWidgets.Container(
          padding: edgeInsetsSmall,
          height: 3.0 * PdfPageFormat.cm,
          width: 3.0 * PdfPageFormat.cm,
          child: icon
      );
  }

  static String cleanType(String text) {
    return text.replaceAll("SuRecordType.", "").toUpperCase();
  }

  static String cleanAlgo(String text) {
    return text.replaceAll("Algo.", "").replaceAll("_", "-").toUpperCase();
  }

  /*
  (177)] Unhandled Exception: UnimplementedError
    E/flutter (14391): #0      _DateFormatPatternField.formatTimeZoneRFC (package:intl/src/intl/date_format_field.dart:650:5)
  E/flutter (14391): #1      _DateFormatPatternField
  String formatTimeZoneId(DateTime date) {
    // TODO(alanknight): implement time zone support
    throw UnimplementedError();
  }*/

  static String timestampToTime(int timestamp) {
    var dateTime = DateTime.fromMicrosecondsSinceEpoch(timestamp*1000);
    // return DateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ").format(dateTime);
    return DateFormat("yyyy-MM-dd HH:mm:ss").format(dateTime);
  }

    static pdfWidgets.Widget blackText(String text, double fontSize) {
    return pdfWidgets.Text(text,
        style: pdfWidgets.TextStyle(
          fontSize: fontSize,
          color: PdfColors.black,
        )
    );
  }

  static pdfWidgets.Widget suRecordMetaDataWidget(SuRecord suRecord) {
    return pdfWidgets.Container(
        padding: edgeInsetsSmall,
        color: PdfColors.grey100,
        child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              pdfWidgets.Row(
                  children: [
                    blackText("Record type: ", FontSizeXL),
                    blackText(cleanType(suRecord.type.toString()), FontSizeXL)
                  ]
              ),
            ]
        )
    );
  }

  static List<pdfWidgets.Widget> VerificationsWidget(SuRecord suRecord) {
    var verifications = suRecord.verifications;
    var result = [for (var verification in verifications) verificartionWidget(verification)];
    return result;
  }

  static pdfWidgets.Widget verificartionWidget(Verification verification) {
    return pdfWidgets.Container(
        padding: edgeInsetsSmall,
        color: PdfColors.grey200,
        child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              pdfWidgets.Row(
                  children: [
                    blackText("Verification", FontSizeXL),
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText(" ", FontSizeS)
                  ]
              ),
              pdfWidgets.Row(
                children: [
                  blackText("URI: ", FontSizeM),
                  blackText(verification.uri, FontSizeM)
                ],
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Timestamp: ", FontSizeM),
                    blackText(timestampToTime(verification.timestamp), FontSizeM)
                  ]
              ),
              ...fingerprintWidgets(verification.fingerprints),
              pdfWidgets.Container(
                // padding: const pdfWidgets.EdgeInsets.only(left: 2.0, top: 5.0, bottom: 5.0, right: 2.0),
                color: PdfColors.grey200,
              ),
              confirmationWidget(verification.confirmation, verification.uri)
            ]
        )
    );
  }

  static List<pdfWidgets.Widget> fingerprintWidgets(List<Fingerprint> fingerprints) {
    var result = [for (var fingerprint in fingerprints) fingerprintWidget(fingerprint)];
    return result;
  }

  static pdfWidgets.Widget fingerprintWidget(Fingerprint fingerprint) {
    return pdfWidgets.Container(
        padding: edgeInsetsSmall,
        color: PdfColors.grey200,
        child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              pdfWidgets.Row(
                  children: [
                    blackText("Fingerprint", FontSizeM)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText(" ", FontSizeXS)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Algorithm: ", FontSizeS),
                    blackText(cleanAlgo(fingerprint.algo.toString()), FontSizeS)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Value: ", FontSizeS),
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText(fingerprint.value, FontSizeXS)
                  ]
              ),
            ]
        )
    );
  }


  static pdfWidgets.Widget confirmationWidget(Confirmation confirmation, String uri) {
    return pdfWidgets.Container(
        padding: edgeInsetsSmall,
        color: PdfColors.grey200,
        child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              pdfWidgets.Row(
                  children: [
                    blackText("Confirmation:", FontSizeM)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText(" ", FontSizeXS)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Timestamp: ", FontSizeS),
                    blackText(timestampToTime(confirmation.timestamp), FontSizeS)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Algorithm: ", FontSizeS),
                    blackText(cleanAlgo(confirmation.algo.toString()), FontSizeS)
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText("Value: ", FontSizeS),
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                    blackText(confirmation.value.toString(), FontSizeXS)
                  ]
              ),
              qrCodeWithUrlWidget(webCheckUrlPrefix + confirmation.value.toString(), uri)
            ]
        )
    );
  }


  static pdfWidgets.Widget qrCodeWithUrlWidget(String data, String uri) {

    if( "local".compareTo(uri) != 0) {
      return pdfWidgets.Column(
        mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
        crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
        children: [
          pdfWidgets.Container(
            padding: const pdfWidgets.EdgeInsets.only(left: 10.0, top: 10.0, bottom: 3.0, right: 10.0),
            height: 3.0 * PdfPageFormat.cm,
            width: 3.0 * PdfPageFormat.cm,
            child: pdfWidgets.BarcodeWidget(
              barcode: pdfWidgets.Barcode.qrCode(
                  typeNumber: 10, // typeNumber >= 1 && typeNumber <= 40
                  errorCorrectLevel: pdfWidgets.BarcodeQRCorrectionLevel.medium),
              data: data,
            ),
          ),
          pdfWidgets.Container(
            alignment: pdfWidgets.Alignment.center,
            child: blackText(data, FontSizeXXS),
          )
        ],
      );
    } else {
      return pdfWidgets.Column(
          mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
          crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
          children: [
            blackText("", FontSizeXS),
            blackText("local confirmations can ", FontSizeS),
            blackText("not be checked online", FontSizeS),
          ]
      );
      }
  }

  // ===================================================

  static pdfWidgets.Widget blanko(SuRecord suRecord) {
    pdfWidgets.Container(
        padding: const pdfWidgets.EdgeInsets.only(left: 0.0, top: 0.0, bottom: 5.0, right: 0.0),
        child: pdfWidgets.Column(
            mainAxisAlignment: pdfWidgets.MainAxisAlignment.start,
            crossAxisAlignment: pdfWidgets.CrossAxisAlignment.center,
            children: [
              pdfWidgets.Row(
                  children: [
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                  ]
              ),
              pdfWidgets.Row(
                  children: [
                  ]
              ),
            ]
        )
    );
    return null;
  }
}