/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:convert';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:logging/logging.dart';
import 'package:path_provider/path_provider.dart';
import 'package:surec_app/cache/fs_audio_cache.dart';
import 'package:surec_app/cache/cache_holder.dart';
import 'package:surec_app/cache/fs_cache.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_blocutil.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/views/detail/read_only_tags.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'dart:async';
import 'dart:io';
import 'package:flutter_sound/flutter_sound.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:surec_app/generated/l10n.dart';

typedef OnCaptureDone = Function(BlocSuRecord result);
typedef OnCaptureError = Function(SuRecordFailureCubitStateValue error);

/// The audio capture view captures audio records.
/// It needs [SuRecordCubit] to save the record.
class AudioView extends StatefulWidget {
  static final log = Logger('AudioView');

  final OnCaptureDone _onCaptureDone;
  final OnCaptureError _onCaptureError;

  final String saveButtonKey = "save_icon_text_button";
  final String inputTextFieldKey = "audio_content_textfield";

  const AudioView(
      {@required OnCaptureDone onCaptureDone,
      OnCaptureError onCaptureError,
      Key key})
      : this._onCaptureDone = onCaptureDone,
        this._onCaptureError = onCaptureError,
        super(key: key);

  @override
  _AudioViewState createState() => _AudioViewState();
}

typedef _Fn = void Function();

class _AudioViewState extends State<AudioView> {
  static final log = Logger('_AudioViewState');

  FlutterSoundRecorder _flutterSoundRecorder = FlutterSoundRecorder();
  bool _flutterSoundRecorderInitialized = false;

  final String _transactionIdPrefix = "audio_capture_transaction_";

  String _recordSaveTransaction;
  bool _runningRecording = false;
  bool _processingRecording = false;

  FSAudioCache _audioCache = CacheHolder().getAudioCache();

  String _tempPathWithFilename;

  @override
  void initState() {
    // Do not access FlutterSoundRecorder before the completion of the Future ~ _mRecorderIsInited == true

    prepareRecorder().then((value) {
      setState(() {
        _flutterSoundRecorderInitialized = true;
      });
    });
    super.initState();
  }

  @override
  void dispose() async {
    if (!_flutterSoundRecorder.isStopped) {
      await _flutterSoundRecorder.stopRecorder();
    }
    _flutterSoundRecorder.closeAudioSession();
    _flutterSoundRecorder = null;
    super.dispose();
  }

  void removeTempfile() {
    FSAudioCache _audioCache = CacheHolder().getAudioCache();
    var _tempPathWithFilename = _audioCache.getTempItemCapturePathWithName();
    var _tempFile = File(_tempPathWithFilename);
    try {
      if (_tempFile.existsSync()) {
        _tempFile.delete();
      }
    } catch (e, s) {
      log.info("delete error");
    }
  }

  Future<void> prepareRecorder() async {
    var status = await Permission.microphone.request();
    if (status != PermissionStatus.granted) {
      throw RecordingPermissionException('Microphone permission not granted');
    }

    removeTempfile();

    await _flutterSoundRecorder.openAudioSession();
    _flutterSoundRecorderInitialized = true;
  }

  Future<void> record() async {
    assert(_flutterSoundRecorderInitialized);

    assert(_flutterSoundRecorderInitialized && _flutterSoundRecorder.isStopped);

    _recordSaveTransaction = _generateTransaction();

    var timestamp = FSCache.timestamp;
    Directory tempDir = await getTemporaryDirectory();
    _tempPathWithFilename = '${tempDir.path}/$timestamp.aac.tmp';

    await _flutterSoundRecorder.startRecorder(
      toFile: _tempPathWithFilename,
      codec: Codec.aacADTS,
    );
    setState(() {
      _runningRecording = true;
    });
  }

  Future<void> stopRecorder() async {
    _processingRecording = true;
    _runningRecording = false;
    if (!_flutterSoundRecorder.isStopped) {
      await _flutterSoundRecorder.stopRecorder();
    }
    createAudioRecord(context, _recordSaveTransaction);
  }

  _Fn getRecorderFn() {
    if (!_flutterSoundRecorderInitialized) {
      return null;
    }
    return _flutterSoundRecorder.isStopped
        ? record
        : () {
            stopRecorder().then((value) => setState(() {}));
          };
  }

  @override
  Widget build(BuildContext context) {
    return BlocListener<SuRecordCubit, SuRecordCubitState>(
        listener: (context, suRecordCubitState) {
          if (suRecordCubitState is SuRecordFailureCubitState &&
              suRecordCubitState.transaction == _recordSaveTransaction) {
            widget._onCaptureError(suRecordCubitState.value);
          } else if (suRecordCubitState is SuRecordSavedCubitState &&
              suRecordCubitState.transaction == _recordSaveTransaction) {
            var suRecord = suRecordCubitState.blocSuRecords
                .firstWhere((record) => record.id == suRecordCubitState.id);
            var cacheHolder = CacheHolder();
            if (cacheHolder.isInitializedAudioCache()) {
              var audioCache = cacheHolder.getAudioCache();
              var cacheEntryName =
                  BlocSuRecordUtil.generateCacheEntryName(suRecord);
              audioCache.cacheItemFileFromPath(
                  _tempPathWithFilename, cacheEntryName);
            }
            widget._onCaptureDone(suRecord);
          }
        },
        child: Container(
          width: double.infinity,
          child: _buildRecordUI(),
          color: Theme.of(context).colorScheme.surface,
        ));
  }

  Widget _buildRecordUI() {
    return Column(mainAxisAlignment: MainAxisAlignment.center, children: [
      Text((_flutterSoundRecorder.isRecording | _processingRecording)
          ? S.of(context).audioViewRecordingInProgressLabel
          : S.of(context).audioViewRecordingIsReadyLabel),
      ActionButton(
        icon: (_flutterSoundRecorder.isRecording | _processingRecording)
            ? LayoutConfig.audioAndVideoStopIcon
            : LayoutConfig.audioAndVideoPlayIcon,
        label: (_flutterSoundRecorder.isRecording | _processingRecording)
            ? S.of(context).audioViewStopButtonLabel
            : S.of(context).audioViewRecordButtonLabel,
        onPressed: getRecorderFn(),
      ),
    ]);
  }

  void createAudioRecord(BuildContext context, String transaction) {
    var suRecordCubit = BlocProvider.of<SuRecordCubit>(context);
    var recordType = SuRecordType.aac;
    suRecordCubit.saveFile(_tempPathWithFilename, recordType,
        ReadOnlyTags.getReadOnlyTagsOfType(recordType), [], transaction);
  }

  String _generateTransaction() {
    return "$_transactionIdPrefix ${DateTime.now().millisecondsSinceEpoch.toString()}";
  }
}
