/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

// This class uses code from https://pub.dev/packages/camera/example
//
// Copyright 2019 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
// https://github.com/flutter/plugins/blob/master/packages/camera/camera/LICENSE

import 'dart:async';

import 'package:camera/camera.dart';
import 'package:camera_platform_interface/camera_platform_interface.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:logging/logging.dart';
import 'package:surec_app/cache/cache_holder.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_blocutil.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/surecord/surecord_exception.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:surec_app/ui/widgets/card_container.dart';
import 'package:video_player/video_player.dart';
import 'package:surec_app/generated/l10n.dart';
import 'dart:io';
import 'package:exif/exif.dart';
import 'package:image/image.dart' as img;

import '../../system_config.dart';
import 'detail/read_only_tags.dart';


enum CamViewMode { PHOTO, VIDEO }

typedef OnCaptureDone = Function(BlocSuRecord result);
typedef OnCaptureError = Function(SuRecordFailureCubitStateValue error);

class CamView extends StatefulWidget {
  static final log = Logger('CamView');

  final OnCaptureDone _onCaptureDone;
  final OnCaptureError _onCaptureError;
  final CamViewMode _camViewMode;


  const CamView(
      {@required CamViewMode camViewMode,
        @required OnCaptureDone onCaptureDone,
        OnCaptureError onCaptureError,
        Key key})
      : this._camViewMode = camViewMode,
        this._onCaptureDone = onCaptureDone,
        this._onCaptureError = onCaptureError,
        super(key: key);

  @override
  _CamViewState createState() {
    return _CamViewState();
  }
}


class _CamViewState extends State<CamView> with WidgetsBindingObserver, TickerProviderStateMixin {
  static final log = Logger('_CamViewState');

  final  _transactionIdPrefixVideo  = "video_capture_transaction_";
  final  _transactionIdPrefixPhoto  = "photo_capture_transaction_";
  
  CameraController controller;
  XFile imageFile;
  XFile videoFile;
  VideoPlayerController videoController;
  VoidCallback videoPlayerListener;
  bool _enableAudio = true;
  double _minAvailableExposureOffset = 0.0;
  double _maxAvailableExposureOffset = 0.0;
  double _currentExposureOffset = 0.0;
  AnimationController _flashModeControlRowAnimationController;
  Animation<double> _flashModeControlRowAnimation;
  AnimationController _exposureModeControlRowAnimationController;
  Animation<double> _exposureModeControlRowAnimation;
  int _pointers = 0;   // Counting pointers (number of user fingers on screen)
  double _minAvailableZoom;
  double _maxAvailableZoom;
  double _currentScale = 1.0;
  double _baseScale = 1.0;
  CameraLensDirection _cameraLensDirection;
  String _transactionPrefix;
  String _transaction;


  @override
  void initState() {
    super.initState();
    WidgetsBinding.instance.addObserver(this);
    _flashModeControlRowAnimationController = AnimationController(
      duration: const Duration(milliseconds: 300),
      vsync: this,
    );
    _flashModeControlRowAnimation = CurvedAnimation(
      parent: _flashModeControlRowAnimationController,
      curve: Curves.easeInCubic,
    );
    _exposureModeControlRowAnimationController = AnimationController(
      duration: const Duration(milliseconds: 300),
      vsync: this,
    );
    _exposureModeControlRowAnimation = CurvedAnimation(
      parent: _exposureModeControlRowAnimationController,
      curve: Curves.easeInCubic,
    );

    _cameraLensDirection = _initialCameraLensDirection();
    var cameraDescription = SystemConfig().cameraDescriptions.firstWhere(
            (element) => element.lensDirection == _cameraLensDirection);
    onNewCameraSelected(cameraDescription);

    _transactionPrefix = widget._camViewMode ==
        CamViewMode. PHOTO ? _transactionIdPrefixPhoto : _transactionIdPrefixVideo;
  }


  @override
  Widget build(BuildContext context) {
    return BlocListener<SuRecordCubit, SuRecordCubitState>(
        listener: (context, suRecordCubitState) {
          if (suRecordCubitState is SuRecordFailureCubitState &&
              suRecordCubitState.transaction == _transaction) {
            final message = "SuRecordFailureCubitState in build() for transaction $_transaction";
            log.severe("$message");
            widget._onCaptureError(suRecordCubitState.value);
          } else if (suRecordCubitState is SuRecordSavedCubitState
              && suRecordCubitState.transaction == _transaction) {

            var blocSuRecord = suRecordCubitState.blocSuRecords.firstWhere(
                    (record) => record.id == suRecordCubitState.id);

            if(_transaction.startsWith(_transactionIdPrefixPhoto)) {
              var cacheHolder = CacheHolder(); // TODO: DI?
              if (cacheHolder.isInitializedImageAndVideoCache()) {
                var cacheEntryName = BlocSuRecordUtil.generateCacheEntryName(blocSuRecord);
                cacheHolder.getImageAndVideoCache().cacheImageFileFromPath(imageFile.path, cacheEntryName);
              }
              widget._onCaptureDone(blocSuRecord);
            } else if(_transaction.startsWith(_transactionIdPrefixVideo)) {
              var cacheHolder = CacheHolder(); // TODO: DI?
              if (cacheHolder.isInitializedImageAndVideoCache()) {
                var cacheEntryName = BlocSuRecordUtil.generateCacheEntryName(blocSuRecord);
                cacheHolder.getImageAndVideoCache().cacheVideoFileFromPath(videoFile.path, cacheEntryName);
              }
              widget._onCaptureDone(blocSuRecord);

            } else {
              // invalid transaction code
            }
          }
        },
        child: Container(
          color: Theme.of(context).colorScheme.background,
          child: widget._camViewMode == CamViewMode.PHOTO ?
            _buildPhotoUI(context) : _buildVideoUI(context)
        )
    );
  }


  Widget _buildPhotoUI(BuildContext context) {
    return Column(
      children: <Widget>[
        Expanded(
          child: Container(
            child: Padding(
              padding: const EdgeInsets.all(1.0), // 1
              child: Center(
                child: _cameraPreviewWidget(),
              ),
            ),
            decoration: BoxDecoration(
              color: Colors.black,
              border: Border.all(
                color: controller != null && controller.value.isRecordingVideo ? Colors.redAccent : Colors.black,
                width: 2.0, // 3
              ),
            ),
          ),
        ),
        
        _modeControlRowWidget(),
        _captureControlRowWidget(),
      ],
    );
  }

  Widget _buildVideoUI(BuildContext context) {
    return Column(
      children: <Widget>[
        Expanded(
          child: Container(
            child: Padding(
              padding: const EdgeInsets.all(1.0), // 1
              child: Center(
                child: _cameraPreviewWidget(),
              ),
            ),
            decoration: BoxDecoration(
              color: Colors.black,
              border: Border.all(
                color: controller != null && controller.value.isRecordingVideo ? Colors.redAccent : Colors.black,
                width: 2.0, // 3
              ),
            ),
          ),
        ),
        _modeControlRowWidget(),
        _captureControlRowWidget(),
      ],
    );
  }



  /// Display the preview from the camera (or a message if the preview is not available).
  Widget _cameraPreviewWidget() {
    if (controller == null || !controller.value.isInitialized) {
      return Text(
        S.of(context).camViewNotInitializedLabel,
        style: TextStyle(
          color: Colors.white,
          fontSize: 24.0,
          fontWeight: FontWeight.w900,
        ),
      );
    } else {
      return AspectRatio(
        aspectRatio: controller.value.aspectRatio,
        child: Listener(
          onPointerDown: (_) => _pointers++,
          onPointerUp: (_) => _pointers--,
          child: LayoutBuilder(builder: (BuildContext context, BoxConstraints constraints) {
            return GestureDetector(
              onScaleStart: _handleScaleStart,
              onScaleUpdate: _handleScaleUpdate,
              onTapDown: (details) => onViewFinderTap(details, constraints),
              child: CameraPreview(controller),
            );
          }),
        ),
      );
    }
  }


  void _handleScaleStart(ScaleStartDetails details) {
    _baseScale = _currentScale;
  }


  Future<void> _handleScaleUpdate(ScaleUpdateDetails details) async {
    // When there are not exactly two fingers on screen don't scale
    if (_pointers != 2) {
      return;
    }
    _currentScale = (_baseScale * details.scale).clamp(_minAvailableZoom, _maxAvailableZoom);

    await controller.setZoomLevel(_currentScale);
  }


  /// Display a bar with buttons to change the flash and exposure modes
  Widget _modeControlRowWidget() {
    return CardContainer(      
          child: Column(
        children: [
          Row(
            mainAxisAlignment: MainAxisAlignment.spaceEvenly,
            mainAxisSize: MainAxisSize.max,
            children: <Widget>[
              IconButton(
                icon: Icon(Icons.flash_on),
                color: Theme.of(context).colorScheme.primaryVariant,
                onPressed: controller != null ? onFlashModeButtonPressed : null,
              ),
              IconButton(
                icon: Icon(Icons.exposure),
                color: Theme.of(context).colorScheme.primaryVariant,
                onPressed: controller != null ? onExposureModeButtonPressed : null,
              ),
              _cameraToggleButtonRowWidget(),
            ],
          ),
          _flashModeControlRowWidget(),
          _exposureModeControlRowWidget(),
        ],
      ),
    );
  }


  final cameraLensDirectionOrder = [
    CameraLensDirection.back,
    CameraLensDirection.front,
    CameraLensDirection.external
  ];


  /// select CameraLensDirection from avaliable CameraLensDirection with priority back, front, external
  CameraLensDirection _initialCameraLensDirection() {
    if (SystemConfig().cameraDescriptions.isEmpty) {
      final message = "no camera found";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    } else {
      var availableCameraLensDirections = SystemConfig().cameraDescriptions.map(
              (cameraDescription) => (cameraDescription.lensDirection)).toSet();
      for (var cameraLensDirection in cameraLensDirectionOrder) {
        if (availableCameraLensDirections.contains(cameraLensDirection)) {
          return cameraLensDirection;
        }
      }
      final message = "no valid found";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    }
  }


  CameraToggleButtonsLensDirectionConfig _cameraToggleButtonsLensDirectionConfig(List<CameraDescription> cameraDescriptions) {
    List<bool> selected = [];
    List<Icon> icons = [];
    List<CameraLensDirection> orderedCameraLensDirections = [];

    var availableCameraLensDirections = cameraDescriptions.map(
            (cameraDescription) => (cameraDescription.lensDirection)).toSet();

    for (var cameraLensDirection in cameraLensDirectionOrder) {
      if (availableCameraLensDirections.contains(cameraLensDirection)) {
        icons.add(Icon(getCameraLensIcon(cameraLensDirection)));
        orderedCameraLensDirections.add(cameraLensDirection);
        if (_cameraLensDirection == cameraLensDirection) {
          selected.add(true);
        } else {
          selected.add(false);
        }
      }
    }
    return CameraToggleButtonsLensDirectionConfig(selected, icons, orderedCameraLensDirections);
  }


  /// Display a row of toggle to select the camera (or a message if no camera is available).
  Widget _cameraToggleButtonRowWidget() {
    final List<Widget> toggleButtons = <Widget>[];

    if (SystemConfig().cameraDescriptions.isEmpty) {
      final message = "no camera found";
      log.severe("$message");
      throw SuRecordException(SuRecordExceptionCode.ERROR, explanation: message);
    } else {
      var buildCameraLensDirectionToggleButtonsConfig =
      _cameraToggleButtonsLensDirectionConfig(SystemConfig().cameraDescriptions);

      return ToggleButtons(
        children: buildCameraLensDirectionToggleButtonsConfig.icons,
        isSelected: buildCameraLensDirectionToggleButtonsConfig.selected,
        onPressed: controller != null  && !controller.value.isRecordingVideo
            ? (int index) {setState(() {
                var cameraLensDirection = buildCameraLensDirectionToggleButtonsConfig.orderedCameraLensDirections[index];
                if(_cameraLensDirection != cameraLensDirection) {
                  _cameraLensDirection = cameraLensDirection;
                var cameraDescription = SystemConfig().cameraDescriptions.firstWhere(
                    (element) => element.lensDirection == cameraLensDirection);
                onNewCameraSelected(cameraDescription);
                }
            });}
            : null,
      );
    }
  }


  Widget _flashModeControlRowWidget() {
    return SizeTransition(
      sizeFactor: _flashModeControlRowAnimation,
      child: ClipRect(
        child: Row(
          mainAxisAlignment: MainAxisAlignment.spaceEvenly,
          mainAxisSize: MainAxisSize.max,
          children: [
            IconButton(
              icon: Icon(Icons.flash_off),
              color: controller?.value?.flashMode == FlashMode.off ? Colors.orange : Colors.blue,
              onPressed: controller != null ? () => onSetFlashModeButtonPressed(FlashMode.off) : null,
            ),
            IconButton(
              icon: Icon(Icons.flash_auto),
              color: controller?.value?.flashMode == FlashMode.auto ? Colors.orange : Colors.blue,
              onPressed: controller != null && controller.value.isRecordingVideo
                  ? null
                  : () => onSetFlashModeButtonPressed(FlashMode.auto),
            ),
            IconButton(
              icon: Icon(Icons.flash_on),
              color: controller?.value?.flashMode == FlashMode.always ? Colors.orange : Colors.blue,
              onPressed: controller != null && controller.value.isRecordingVideo
                  ? null
                  : () => onSetFlashModeButtonPressed(FlashMode.always),
            ),
            IconButton(
              icon: Icon(Icons.highlight),
              color: controller?.value?.flashMode == FlashMode.torch ? Colors.orange : Colors.blue,
              onPressed: controller != null ? () => onSetFlashModeButtonPressed(FlashMode.torch) : null,
            ),
          ],
        ),
      ),
    );
  }


  Widget _exposureModeControlRowWidget() {
    return SizeTransition(
      sizeFactor: _exposureModeControlRowAnimation,
      child: ClipRect(
        child: Container(
          color: Colors.grey.shade50,
          child: Column(
            children: [
              Center(
                child: Text("Exposure Mode"),
              ),
              Row(
                mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                mainAxisSize: MainAxisSize.max,
                children: [
                  FlatButton(
                    child: Text('AUTO'),
                    textColor: controller?.value?.exposureMode == ExposureMode.auto ? Colors.orange : Colors.blue,
                    onPressed: controller != null ? () => onSetExposureModeButtonPressed(ExposureMode.auto) : null,
                    onLongPress: () {
                      if (controller != null) controller.setExposurePoint(null);
                      showInSnackBar('Resetting exposure point');
                    },
                  ),
                  FlatButton(
                    child: Text('LOCKED'),
                    textColor: controller?.value?.exposureMode == ExposureMode.locked ? Colors.orange : Colors.blue,
                    onPressed: controller != null ? () => onSetExposureModeButtonPressed(ExposureMode.locked) : null,
                  ),
                ],
              ),
              Center(
                child: Text("Exposure Offset"),
              ),
              Row(
                mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                mainAxisSize: MainAxisSize.max,
                children: [
                  Text(_minAvailableExposureOffset.toString()),
                  Slider(
                    value: _currentExposureOffset,
                    min: _minAvailableExposureOffset,
                    max: _maxAvailableExposureOffset,
                    label: _currentExposureOffset.toString(),
                    onChanged: _minAvailableExposureOffset == _maxAvailableExposureOffset ? null : setExposureOffset,
                  ),
                  Text(_maxAvailableExposureOffset.toString()),
                ],
              ),
            ],
          ),
        ),
      ),
    );
  }


  /// Display the control bar with buttons to take pictures and record videos.
  Widget _captureControlRowWidget() {
    var readyToStartCapture = controller != null && controller.value.isInitialized && !controller.value.isRecordingVideo;
    var iconColor = Theme.of(context).colorScheme.primary;
    var iconColorStopCapture = Theme.of(context).colorScheme.error;

    return Padding(
      padding: const EdgeInsets.all(LayoutConfig.doubleWidgetSpacing),
      child:   Row(
            mainAxisAlignment: MainAxisAlignment.spaceEvenly,
            mainAxisSize: MainAxisSize.max,
            children: <Widget>[
              widget._camViewMode == CamViewMode.PHOTO ?
              ActionButton(
                icon: Icons.camera_alt,
                label: S.of(context).camViewTakePhotoButtonLabel,
                onPressed: readyToStartCapture
                    ? onTakePictureButtonPressed
                    : null,
              ) :
              ActionButton(
                icon: readyToStartCapture
                    ? Icons.videocam
                    : Icons.adjust,
                label: readyToStartCapture
                  ? S.of(context).camViewStartRecordingVideoButtonLabel
                  : S.of(context).camViewStopRecordingVideoButtonLabel,
                onPressed: readyToStartCapture
                    ? onVideoRecordButtonPressed
                    : onVideoStopButtonPressed,
              ),
            ],
          ),
    );
  }


  /// Display a row of toggle to select the camera (or a message if no camera is available).
  Widget _cameraTogglesRowWidget() {
    final List<Widget> toggles = <Widget>[];

    if (SystemConfig().cameraDescriptions.isEmpty) {
      return Text(S.of(context).camViewNoCameraFoundErrorMessge);
    } else {
      for (CameraDescription cameraDescription in SystemConfig().cameraDescriptions) {
        toggles.add(
          SizedBox(
            width: 90.0,
            child: RadioListTile<CameraDescription>(
              title: Icon(getCameraLensIcon(cameraDescription.lensDirection)),
              groupValue: controller?.description,
              value: cameraDescription,
              onChanged: controller != null && controller.value.isRecordingVideo ? null : onNewCameraSelected,
            ),
          ),
        );
      }
    }
    return Row(children: toggles);
  }


  String timestamp() => DateTime.now().millisecondsSinceEpoch.toString();


  void onViewFinderTap(TapDownDetails details, BoxConstraints constraints) {
    controller.setExposurePoint(Offset(
      details.localPosition.dx / constraints.maxWidth,
      details.localPosition.dy / constraints.maxHeight,
    ));
  }


  void onNewCameraSelected(CameraDescription cameraDescription) async {
    if (controller != null) {
      await controller.dispose();
    }
    controller = CameraController(
      cameraDescription,
      ResolutionPreset.medium,
      enableAudio: _enableAudio,
    );

    // If the controller is updated then update the UI.
    controller.addListener(() {
      if (mounted) setState(() {});
      if (controller.value.hasError) {
        showInSnackBar('Camera error ${controller.value.errorDescription}');
      }
    });

    try {
      await controller.initialize();
      _minAvailableExposureOffset = await controller.getMinExposureOffset();
      _maxAvailableExposureOffset = await controller.getMaxExposureOffset();
      _maxAvailableZoom = await controller.getMaxZoomLevel();
      _minAvailableZoom = await controller.getMinZoomLevel();
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
    }

    if (mounted) {
      setState(() {});
    }
  }


  void onTakePictureButtonPressed() {
    takePicture().then((XFile file) {
      if (mounted) {
        setState(() {
          imageFile = file;
          videoController?.dispose();
          videoController = null;
        });
        if (file != null) {
          
          changeRotation(file.path);
          
          _transaction = _generateTransaction();
          createSuRecord(file.path, _transaction);
        } else {
          final message = "takePicture() did not deliver file";
          log.severe("$message");
          // TODO: handle error visually

        }
      }});
  }


  void onFlashModeButtonPressed() {
    if (_flashModeControlRowAnimationController.value == 1) {
      _flashModeControlRowAnimationController.reverse();
    } else {
      _flashModeControlRowAnimationController.forward();
      _exposureModeControlRowAnimationController.reverse();
    }
  }


  void onExposureModeButtonPressed() {
    if (_exposureModeControlRowAnimationController.value == 1) {
      _exposureModeControlRowAnimationController.reverse();
    } else {
      _exposureModeControlRowAnimationController.forward();
      _flashModeControlRowAnimationController.reverse();
    }
  }


  void onAudioModeButtonPressed() {
    _enableAudio = !_enableAudio;
    if (controller != null) {
      onNewCameraSelected(controller.description);
    }
  }


  void onSetFlashModeButtonPressed(FlashMode mode) {
    setFlashMode(mode).then((_) {
      if (mounted) setState(() {});
      showInSnackBar('${S.of(context).camViewSetFlashModeButtonSnackBarMessage} ${mode.toString().split('.').last}');
    });
  }

  void onSetExposureModeButtonPressed(ExposureMode mode) {
    setExposureMode(mode).then((_) {
      if (mounted) setState(() {});
      showInSnackBar('${S.of(context).camViewSetExposureModeButtonSnackBarMessage} ${mode.toString().split('.').last}');
    });
  }


  void onVideoRecordButtonPressed() {
    startVideoRecording().then((_) {
      if (mounted) {
        setState( () {} );
      }
      else {
        final message = "startVideoRecording did not resolve to mounted";
        log.severe("$message");
      };
    });
  }

  void onVideoStopButtonPressed() {
    stopVideoRecording().then((file) {
      if (mounted) setState(() {});
      if (file != null) {

        videoFile = file;
        if (file != null) {
          _transaction = _generateTransaction();
          createSuRecord(file.path, _transaction);
        } else {
          final message = "stopVideoRecording() did not deliver file";
          log.severe("$message");
        }
      }
    });
  }


  Future<void> startVideoRecording() async {
    if (!controller.value.isInitialized) {
      showInSnackBar(S.of(context).camViewTakePictureErrorMessge);
      return;
    }

    if (controller.value.isRecordingVideo) {
      // A recording is already started, do nothing.
      return;
    }

    try {
      await controller.startVideoRecording();
    } on CameraException catch (e, s) {
      _showCameraException(e, s);

      return;
    }
  }

  Future<XFile> stopVideoRecording() async {
    if (!controller.value.isRecordingVideo) {
      return null;
    }

    try {
      return controller.stopVideoRecording();
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
      return null;
    }
  }


  Future<void> setFlashMode(FlashMode mode) async {
    try {
      await controller.setFlashMode(mode);
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
      rethrow;
    }
  }

  Future<void> setExposureMode(ExposureMode mode) async {
    try {
      await controller.setExposureMode(mode);
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
      rethrow;
    }
  }

  Future<void> setExposureOffset(double offset) async {
    setState(() {
      _currentExposureOffset = offset;
    });
    try {
      offset = await controller.setExposureOffset(offset);
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
      rethrow;
    }
  }


  Future<XFile> takePicture() async { // called by onTakePictureButtonPressed()
    if (!controller.value.isInitialized) {
      showInSnackBar(S.of(context).camViewTakePictureErrorMessge);
      return null;
    }

    if (controller.value.isTakingPicture) {
      // A capture is already pending, do nothing.
      return null;
    }

    try {
      XFile file = await controller.takePicture();
      return file;
    } on CameraException catch (e, s) {
      _showCameraException(e, s);
      return null;
    }
  }

  void showInSnackBar(String message) {
    Scaffold.of(context).showSnackBar(SnackBar(content: Text(message)));
  }

  void _showCameraException(CameraException exception, StackTrace stackTrace) {
    final exceptionStr = exception.toString();
    final stackTraceStr = stackTrace.toString();
    final message = "CameraException happened";
    log.severe("$message\n$exceptionStr\n$stackTraceStr");
  }


  void createSuRecord(String pathWithName, String transaction) {

    if(pathWithName == null) {
      final message = "createSuRecord() can not create surecord from nonexisting file";
      log.severe(message);
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    }

    var suRecordType;
    if(pathWithName.toLowerCase().endsWith(".png")) {
      suRecordType = SuRecordType.png;
    } else if(pathWithName.toLowerCase().endsWith(".jpg")) {
      suRecordType = SuRecordType.jpg;
    } else if(pathWithName.toLowerCase().endsWith(".mp4")) {
      suRecordType = SuRecordType.mp4;
    } else {
      final message = "createSuRecord() can not create surecord from file with unkown type: $pathWithName";
      log.severe(message);
      throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
    }
    List<String> tags =  ReadOnlyTags.getReadOnlyTagsOfType(suRecordType);
    List<String> refs = [];

    var suRecordCubit = BlocProvider.of<SuRecordCubit>(context);
    suRecordCubit.saveFile(
        pathWithName,
        suRecordType,
        tags,
        refs,        
        transaction
    );
  }

  String _generateTransaction() {
    return "$_transactionPrefix${DateTime.now().millisecondsSinceEpoch.toString()}";
  }


  void changeRotation(String imagePath) {
    final originalFile = File(imagePath);
    List<int> imageBytes = originalFile.readAsBytesSync();

    final originalImage = img.decodeImage(imageBytes);

    final height = originalImage.height;
    final width = originalImage.width;

    if (height >= width) {
      return; // originalFile;
    }

    img.Image fixedImage;

    fixedImage = img.copyRotate(originalImage, 90);

    final fixedFile = originalFile.writeAsBytesSync(img.encodeJpg(fixedImage));
  }
}



class CameraToggleButtonsLensDirectionConfig {
  List<bool> selected = [];
  List<Icon> icons = [];
  List<CameraLensDirection> orderedCameraLensDirections = [];

  CameraToggleButtonsLensDirectionConfig(this.selected, this.icons, this.orderedCameraLensDirections);
}


/// Returns a suitable camera icon for [direction].
IconData getCameraLensIcon(CameraLensDirection direction) {
  switch (direction) {
    case CameraLensDirection.back:
      return Icons.camera_rear;
    case CameraLensDirection.front:
      return Icons.camera_front;
    case CameraLensDirection.external:
      return Icons.camera;
  }
  final message = "Unknown lens direction': $direction";
  throw SuRecordException(SuRecordExceptionCode.INVALID_ARGUMENT, explanation: message);
}