/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/ui/views/detail/read_only_tags.dart';
import 'package:surec_app/ui/views/detail/tag_item.dart';

part 'detail_tag_state.dart';

class DetailTagsCubit extends Cubit<DetailTagState> {
  
  final String _transactionPrefix = "detail_tag_transaction_";
  
  final SuRecordCubit _suRecordCubit;
  
  StreamSubscription _subscription;

  DetailTagsCubit(this._suRecordCubit, String recordId)
    : super(InitialDetailTagState(recordId)){

      _subscription = _suRecordCubit.listen((_suRecordCubit) { 
        if(state is !WorkingDetailTagState){ 
          return; 
        }    
        var currentState = state as WorkingDetailTagState;

        if( _suRecordCubit is SuRecordFindByIDCubitState
          && currentState.transaction == _suRecordCubit.transaction){

          emit(FindAllAvailableTagsDetailTagState(currentState.recordId, _suRecordCubit.blocSuRecord.tags));
        }
        else if( _suRecordCubit is SuRecordCountCubitState
          && currentState.transaction == _suRecordCubit.transaction){ 

          var recordTagItems = _createRecordTagItemList(currentState.recordTags);
          var availableTagItems = _createAvailabelTagItemList(currentState.recordTags, _suRecordCubit.tags.toList());
          emit(EditDetailTagState(currentState.recordId, recordTagItems, availableTagItems));
        }
        else if( _suRecordCubit is SuRecordSavedCubitState
          && currentState.transaction == _suRecordCubit.transaction){

          emit(DoneDetailTagState());
        }
        else if(_suRecordCubit is SuRecordFailureCubitState
          && currentState.transaction == _suRecordCubit.transaction){
          
          emit(FailureTagState());
        }
      });
    }

  void findRecordTags(){
    assert(state is InitialDetailTagState);
    var currentState = state as InitialDetailTagState;

    var tansaction = createTransaction();
    emit(WorkingDetailTagState(currentState.recordId, tansaction));
    _suRecordCubit.findById(currentState.recordId, tansaction);
  }

  void findAllAvailableTags(){
    assert(state is FindAllAvailableTagsDetailTagState);
    var currentState = state as FindAllAvailableTagsDetailTagState;

    var tansaction = createTransaction();
    emit(WorkingDetailTagState(currentState.recordId, tansaction, recordTags: currentState.recordTags));
    
    _suRecordCubit.count(tansaction);  
  }

  void saveRecordTags(){
    assert(state is EditDetailTagState);
    var currentState = state as EditDetailTagState;

    var transaction = createTransaction();
    var tags = currentState.recordTagItems.map((item) => item.name).toList();

    emit(WorkingDetailTagState.fromEditState(currentState, transaction, recordTags: tags));
    _suRecordCubit.saveMetaTagsRefs(currentState.recordId, tags, const[], transaction);
  }

  void deleteRecordTag(String tagName){
    assert(state is EditDetailTagState);
    var currentState = state as EditDetailTagState;

    var isTagSelected = currentState.availableTagItems
      .where((tag) => tag.name == tagName && tag.isSelected);  

    if(isTagSelected.isNotEmpty){
      List<TagItem> recordTags = [];
      recordTags.addAll(currentState.recordTagItems);
      recordTags.removeWhere((tagItem) => tagItem.name == tagName);

      var availableTags = _diselecteAvailableTag(tagName);

      emit(EditDetailTagState(currentState.recordId, recordTags, availableTags));
    }
  }

  void addAvailableTagToRecordTags(String tagName){
    assert(state is EditDetailTagState);
    var currentState = (state as EditDetailTagState);
    
    var tagExist = currentState.recordTagItems
      .where((tag) => tag.name == tagName);

    if(tagExist.isEmpty){      
      List<TagItem> recordTags = [];
      recordTags.addAll(currentState.recordTagItems);
      recordTags.add(TagItem(tagName));

      var availableTags = _selecteAvailableTag(tagName);

      emit(EditDetailTagState(currentState.recordId, recordTags, availableTags));
    }
  }

  void createNewAvailableTag(String tagName){
    assert(tagName.isNotEmpty);
    
    assert(state is EditDetailTagState);
    var currentState = (state as EditDetailTagState);

    var items = currentState.availableTagItems
      .where((item) => item.name == tagName);

    if(items.isEmpty){
      var availableTags = _addNewTagTo(currentState.availableTagItems, tagName, true);
      var recordTags = _addNewTagTo(currentState.recordTagItems, tagName, false);

      emit(EditDetailTagState(currentState.recordId, recordTags, availableTags));
    }
  }

  List<TagItem> _createRecordTagItemList(List<String> recordTags){
    List<TagItem> recordTagItems = [];
   
    for(var tag in recordTags){
      var item = TagItem(tag);
      recordTagItems.add(item);
    }
    return recordTagItems;
  }

  List<TagItem> _createAvailabelTagItemList(List<String> recordTags, List<String> availableTags){
    List<TagItem> availableTagItems = [];

    for(var tag in availableTags){
      if(ReadOnlyTags.isNotReadOnly(tag)){
        var item = TagItem(tag);
        if( recordTags.contains(tag)){
          item.isSelected = true;
        }
        availableTagItems.add(item);
      }
    }
    return availableTagItems;
  }

  List<TagItem> _addNewTagTo(List<TagItem> tagList, String tagName, bool isAvailableTagList){
    var item = TagItem(
      tagName, 
      isSelected: isAvailableTagList ? true : false
    );

    List<TagItem> newTagList = [];
    newTagList.addAll(tagList);
    newTagList.add(item);   

    return newTagList;
  }

  List<TagItem> _diselecteAvailableTag(String tagName){
    var availableTags = _setSelectedValueFromAvailableTags(tagName, false); 

    return availableTags;
  }

  List<TagItem> _selecteAvailableTag(String tagName){
    var availableTags = _setSelectedValueFromAvailableTags(tagName, true);

    return availableTags;
  }

  List<TagItem> _setSelectedValueFromAvailableTags(String tagName, bool value){
    var currentState = (state as EditDetailTagState);

    List<TagItem> availableTags = [];
    availableTags.addAll(currentState.availableTagItems);
    
    var tag = availableTags.where((tagItem) => tagItem.name == tagName).single;
    tag.isSelected = value;

    return availableTags;
  }

  String createTransaction(){
    return "$_transactionPrefix ${DateTime.now().millisecondsSinceEpoch.toString()}";
  }

  @override 
  Future<void> close(){
    if(_subscription != null){
      _subscription.cancel();
    }
    super.close();
  }
}
