/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/views/detail/detail_tag_cubit.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/ui/views/detail/read_only_tags.dart';
import 'package:surec_app/ui/views/detail/tag_item.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:surec_app/ui/widgets/card_container.dart';
import 'package:surec_app/ui/widgets/loading_progress.dart';
import 'package:surec_app/generated/l10n.dart';

/// Callback definition to handle the on save event for the record [id] with a list of [recordTags].
// typedef OnCaptureTagsDone = void Function(String id, List<String> recordTags);
typedef OnDoneDetailTagCapture = void Function();
typedef OnErrorDetailTagCapture = void Function();

/// The tag view shows a list of all available tags [availableTagNames] and the current record tags [recordTagNames].
/// The user can add new tags via an input text field or arrange tags between the two list via drag&drop or double click.
/// By clicking on the Button the callback [onSave] is called with the current chosen record tags.
class DetailTagView extends StatefulWidget {
  final String _recordId;

  /// on save callback to store tags
  final OnDoneDetailTagCapture _onDoneDetailTag;

  final OnErrorDetailTagCapture _onErrorDetailTag;

  const DetailTagView(
      {@required String recordId,
      @required OnDoneDetailTagCapture onDoneDetailTag,
      OnErrorDetailTagCapture onErrorDetailTag})
      : this._recordId = recordId,
        this._onDoneDetailTag = onDoneDetailTag,
        this._onErrorDetailTag = onErrorDetailTag ?? null;

  @override
  _DetailTagViewState createState() => _DetailTagViewState();
}

class _DetailTagViewState extends State<DetailTagView> {
  Color _isSelectedTagColor;


  /// controller to handle new tag input field
  final TextEditingController _editingController = TextEditingController();

  @override
  void dispose() {
    _editingController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    _isSelectedTagColor = Theme.of(context).colorScheme.background;
    return BlocProvider<DetailTagsCubit>(
      create: (context) => DetailTagsCubit(
          BlocProvider.of<SuRecordCubit>(context), widget._recordId),
      child: BlocConsumer<DetailTagsCubit, DetailTagState>(
        listener: (context, tagState) {
          if (tagState is DoneDetailTagState) {
            widget._onDoneDetailTag();
          } else if (tagState is FailureTagState) {
            // todo error handling
            widget._onErrorDetailTag();
          }
        },
        builder: (context, tagState) {
          var tagCubit = BlocProvider.of<DetailTagsCubit>(context);
          if (tagState is InitialDetailTagState) {
            tagCubit.findRecordTags();
          } else if (tagState is FindAllAvailableTagsDetailTagState) {
            tagCubit.findAllAvailableTags();
          }

          if (tagState is EditDetailTagState) {
            return Padding(
              padding: EdgeInsets.all(LayoutConfig.doubleWidgetSpacing),
              child: Column(
                mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                children: [
                  Expanded(
                    key: Key("available_tag_area"),
                    child: _buildAvailableTags(
                        tags: tagState.availableTagItems,
                        onDoubleTap: (tagName) =>
                            tagCubit.addAvailableTagToRecordTags(tagName),
                        onTagDropped: (tagName) =>
                            tagCubit.deleteRecordTag(tagName),
                        onCreateNewTag: (tagName) {
                          if (ReadOnlyTags.isReadOnly(
                              _editingController.text)) {
                            final snackBar = SnackBar(
                                content: Text( "${S.of(context).detailTagViewInputReadOnlyTagError} '${ReadOnlyTags.READ_ONLY_PREFIX}'"));
                            ScaffoldMessenger.of(context)
                                .showSnackBar(snackBar);
                          } else {
                            tagCubit.createNewAvailableTag(tagName);
                            _editingController.clear();
                            FocusScope.of(context).unfocus();
                          }
                        }),
                  ),
                  // Divider(),
                  Expanded(
                    key: Key("record_tag_area"),
                    child: _buildRecordTags(
                        tags: tagState.recordTagItems,
                        onTagDropped: (tagName) =>
                            tagCubit.addAvailableTagToRecordTags(tagName),
                        onDeleteTag: (tagName) =>
                            tagCubit.deleteRecordTag(tagName)),
                  ),
                  Row(
                      mainAxisAlignment: MainAxisAlignment.spaceAround,
                      children: [
                        ActionButton(
                            icon: LayoutConfig.saveRecordIcon,
                            label: S.of(context).detailTagViewSaveButtonLabel,
                            onPressed: () => tagCubit.saveRecordTags())
                      ]),
                ],
              ),
            );
          } else {
            // if not editState -> show loading progress
            return LoadingProgress();
          }
        },
      ),
    );
  }

  /// creates a tag widget with all available tags
  Widget _buildAvailableTags(
      {List<TagItem> tags,
      Function(String tagName) onDoubleTap,
      Function(String tagName) onTagDropped,
      Function(String tagName) onCreateNewTag}) {
    return _buildTagItems(
        tags: tags,
        onDoubleTap: onDoubleTap,
        onTagDropped: onTagDropped,
        onCreateNewTag: onCreateNewTag,
        onDeleteTag: null);
  }

  /// creates a tag widget with all record tags
  Widget _buildRecordTags(
      {List<TagItem> tags,
      Function(String tagName) onTagDropped,
      Function(String tagName) onDeleteTag}) {
    return _buildTagItems(
        tags: tags,
        onDoubleTap: onDeleteTag,
        onTagDropped: onTagDropped,
        onCreateNewTag: null,
        onDeleteTag: onDeleteTag);
  }

  /// creates tag items that can handle double tap [onDoubleTap], tag dropped [onTagDropped] or delete [onDeleteTag],
  /// when tag action are defined. If [onCreateNewTag] action is defined
  /// an input text field widget is added to the end of the tag list.
  Widget _buildTagItems(
      {List<TagItem> tags,
      Function(String tagName) onDoubleTap,
      Function(String tagName) onTagDropped,
      Function(String tagName) onCreateNewTag,
      Function(String tagName) onDeleteTag}) {
    assert(tags != null, "Error, tags are null");
    List<Widget> items = List.generate(
        tags.length,
        (idx) => _buildTagItem(
            tagItem: tags[idx],
            onDoubleTap: onDoubleTap,
            onDeleteTag: onDeleteTag));

    if (onCreateNewTag != null) {
      var inputTf = TextField(
        keyboardType: TextInputType.text,
        controller: _editingController,
        onSubmitted: (newTagName) => onCreateNewTag(newTagName),
        decoration: InputDecoration(
          hintText: S.of(context).detailTagViewAddTagHintLabel,
        ),
      );

      items.add(inputTf);
    }

    return DragTarget<String>(
      builder: (context, candidateData, rejecteData) {
        return CardContainer(
          alignment: Alignment.topLeft,
          child: Scrollbar(
            child: SingleChildScrollView(
                child: Wrap(spacing: 4.0, runSpacing: 4.0, children: items)),
          ),
        );
      },
      onAccept: (data) => onTagDropped(data),
      onWillAccept: (data) => true,
    );
  }

  Widget _buildTagItem(
      {TagItem tagItem,
      Function(String tagName) onDoubleTap,
      Function(String tagName) onDeleteTag}) {
    Color tagColor;
    // var isReadOnly = _readOnlyTags.contains(tagItem.name);
    var tagItemName = tagItem.name;
    var isReadOnly = ReadOnlyTags.isReadOnly(tagItemName);
    if (isReadOnly) {
      tagItemName = ReadOnlyTags.getTagWithoutPrefix(tagItemName);
      tagColor = Theme.of(context).colorScheme.secondary;
    } else if (tagItem.isSelected) {
      tagColor = _isSelectedTagColor;
    } else {
      tagColor = Theme.of(context).colorScheme.primary;
    }

    var chip = _buildTagAsChip(
        label: tagItemName,
        color: tagColor,
        onDoubleTap: !isReadOnly ? onDoubleTap : null,
        onDeleteTag: !isReadOnly ? onDeleteTag : null);

    var chipWhenDragging = _buildTagAsChip(
        label: tagItemName, color: _isSelectedTagColor, onDeleteTag: null);

    /// tag is not draggable, when the tag is readOnly or already in the record list.
    return isReadOnly || tagItem.isSelected
        ? chip
        : Draggable<String>(
            child: chip,
            feedback: Material(
              color: Color(0x00000000),
              /* material as transparent */
              child: chip,
            ),
            childWhenDragging: chipWhenDragging,
            data: tagItem.name,
          );
  }

  /// create a tag widget with the given [label] and [color] and the defined actions for [onDoubleTap] or [onDeleteTag].
  Widget _buildTagAsChip(
      {String label,
      Color color,
      Function(String tagName) onDoubleTap,
      Function(String tagName) onDeleteTag}) {
    return GestureDetector(
      onDoubleTap: onDoubleTap != null ? () => onDoubleTap(label) : null,
      child: Chip(
          label: Text(label),
          backgroundColor: color,
          onDeleted: onDeleteTag != null ? () => onDeleteTag(label) : null,
          deleteIconColor: Theme.of(context).colorScheme.surface,
          // labelStyle: TextStyle(color: Theme.of(context).colorScheme.surface),
          ),
          
    );
  }
}
