/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:surec_app/cache/cache_holder.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_blocutil.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/util/date_time_formatter_util.dart';
import 'package:surec_app/ui/views/detail/read_only_tags.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:surec_app/ui/widgets/card_container.dart';
import 'package:surec_app/generated/l10n.dart';
import 'package:surec_app/ui/widgets/loading_progress.dart';

typedef OnDoneNoteCapture = Function(BlocSuRecord result);
typedef OnErrorNoteCapture = Function(SuRecordFailureCubitStateValue error);

class NoteView extends StatefulWidget {
  final OnDoneNoteCapture _onCaptureRecordDone;
  final OnErrorNoteCapture _onCaptureRecordError;

  final String saveButtonKey = "save_icon_text_button";
  final String inputTextFieldKey = "note_content_textfield";

  const NoteView(
      {@required OnDoneNoteCapture onDoneNoteCapture,
      OnErrorNoteCapture onErrorNoteCapture,
      Key key})
      : this._onCaptureRecordDone = onDoneNoteCapture,
        this._onCaptureRecordError = onErrorNoteCapture,
        super(key: key);

  @override
  _NoteViewState createState() => _NoteViewState();
}

class _NoteViewState extends State<NoteView> {
  TextEditingController _noteController = TextEditingController();

  /// indicates if the save button is clicked
  final String _transactionIdPrefix = "note_capture_transaction_";

  String _noteTransactionId;

  bool _isSaveClicked = false;

  String _currentDate;

  @override
  void initState() {
    super.initState();

    _currentDate =
        DateTimeFormatterUtil.formatDateTimeToYearMonthDay(DateTime.now());
  }

  @override
  Widget build(BuildContext context) {
    return BlocListener<SuRecordCubit, SuRecordCubitState>(
        listener: (context, suRecordCubitState) {
          if (suRecordCubitState is SuRecordFailureCubitState &&
              suRecordCubitState.transaction == _noteTransactionId) {
            // todo -> error handling
            widget._onCaptureRecordError(suRecordCubitState.value);
          } else if (suRecordCubitState is SuRecordSavedCubitState &&
              suRecordCubitState.transaction == _noteTransactionId) {
            var suRecord = suRecordCubitState.blocSuRecords
                .firstWhere((record) => record.id == suRecordCubitState.id);
            _tryCachingNote(suRecord, _noteController.text);
            widget._onCaptureRecordDone(suRecord);
          }
        },
        child: Column(
          children: [
            CardContainer(
              child: Text(
                _currentDate,
                textScaleFactor: 1.1,
                style: TextStyle(fontWeight: FontWeight.bold),
              ),
              alignment: Alignment.centerLeft,
            ),
            Expanded(
              child: CardContainer(
                child: _isSaveClicked
                    ? Stack(children: [
                        _buildInputField(isEnabled: false),
                        LoadingProgress(),
                      ])
                    : Column(
                        mainAxisAlignment: MainAxisAlignment.start,
                        children: [_buildInputField(isEnabled: true)],
                      ),
              ),
            ),
            Divider(),
            _buildSaveButton(),
          ],
        ));
  }

  @override
  void dispose() {
    _noteController.dispose();
    super.dispose();
  }

  void _tryCachingNote(BlocSuRecord blocSuRecord, String text) {
    var cacheHolder = CacheHolder();
    if (cacheHolder.isInitializedNoteCache()) {
      var cacheEntryName =
          BlocSuRecordUtil.generateCacheEntryName(blocSuRecord);
      cacheHolder.getNoteCache().cacheText(text, cacheEntryName);
    }
  }

  /// creates a widget with available note actions.
  Widget _buildSaveButton() {
    return Container(
      padding: EdgeInsets.only(bottom: 16.0),
      child: Row(mainAxisAlignment: MainAxisAlignment.spaceAround, children: [
        ActionButton(
          icon: LayoutConfig.saveRecordIcon,
          label: S.of(context).noteViewSaveButtonLabel,
          onPressed: () {
            if (_noteController.text.isEmpty) {
              final snackBar = SnackBar(
                  content:
                      Text(S.of(context).noteViewSaveEmptyNoteErrorMessage));
              ScaffoldMessenger.of(context).showSnackBar(snackBar);
            } else {
              FocusScope.of(context).unfocus(); // hide keyboard

              _noteTransactionId = _createTransactionId();
              createNoteRecord(context, _noteTransactionId);
              setState(() {
                _isSaveClicked = true;
              });
            }
          },
        )
      ]),
    );
  }

  /// creates an infinity, scrollable input text field that could be enabled/disabled via [isEnabled].
  Widget _buildInputField({bool isEnabled = true}) {
    return Expanded(
      child: Scrollbar(
        child: SingleChildScrollView(
          scrollDirection: Axis.vertical,
          padding: const EdgeInsets.all(LayoutConfig.doubleWidgetSpacing),
          child: TextField(
            key: Key(widget.inputTextFieldKey),
            cursorColor: Theme.of(context).colorScheme.primaryVariant,
            enabled: isEnabled,
            controller: _noteController,
            autocorrect: true,
            keyboardType: TextInputType.multiline,
            maxLines: null,
            // null = endless lines, starts with one
            decoration: InputDecoration(
                hintText: S.of(context).noteViewHintNoteText,
                focusedBorder: UnderlineInputBorder(
                    borderSide: BorderSide(
                        color: Theme.of(context).colorScheme.primaryVariant))),
          ),
        ),
      ),
    );
  }

  void createNoteRecord(BuildContext context, String transaction) {
    var suRecordCubit = BlocProvider.of<SuRecordCubit>(context);

    var data = utf8.encode(_noteController.text);
    var recordType = SuRecordType.txt;

    suRecordCubit.saveMem(data, recordType,
        ReadOnlyTags.getReadOnlyTagsOfType(recordType), [], transaction);
  }

  String _createTransactionId() {
    return "$_transactionIdPrefix ${DateTime.now().millisecondsSinceEpoch.toString()}";
  }
}
