/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:flutter/material.dart';
import 'package:flutter_sound/flutter_sound.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:surec_app/ui/widgets/loading_progress.dart';
import 'package:surec_app/generated/l10n.dart';

class AudioRecordDataPresentation extends StatefulWidget {
  final String _suRecordDataFile;

  AudioRecordDataPresentation({@required String suRecordDataAsFilePath})
      : _suRecordDataFile = suRecordDataAsFilePath;

  @override
  _AudioRecordDataPresentationState createState() =>
      _AudioRecordDataPresentationState();
}

class _AudioRecordDataPresentationState
    extends State<AudioRecordDataPresentation> {

  final double _textScaleSize = 1.2;

  FlutterSoundPlayer _audioPlayer = FlutterSoundPlayer();

  Duration _maxAudioTrackDuration = Duration(seconds: 0);
  Duration _positionAudioTrack = Duration(seconds: 0);

  double _positionAudioTrackSlider = 0.0;

  bool _isPlayerInitialized = false;

  bool _isUpdatedMaxTrackDuration = false;

  @override
  void initState() {
    super.initState();
    // Be careful : openAudioSession return a Future.
    // Do not access your FlutterSoundPlayer or FlutterSoundRecorder before the completion of the Future
    _audioPlayer.openAudioSession().then((value) {      
      _audioPlayer.setSubscriptionDuration(Duration(milliseconds: 100));
      _audioPlayer.onProgress.listen((event) {
        setState(() {
          _maxAudioTrackDuration = event.duration;
          _positionAudioTrack = event.position;

          var slider = (_positionAudioTrack.inMilliseconds /
                  _maxAudioTrackDuration.inMilliseconds)
              .toDouble();
          _positionAudioTrackSlider = slider;
        });
      });
      setState(() {
        _isPlayerInitialized = true;       
      });
    });
  }

  @override
  Widget build(BuildContext context) {
    // workaround to get the audio track length
    if(_isPlayerInitialized && !_isUpdatedMaxTrackDuration){
      _isUpdatedMaxTrackDuration = true;
      play().then( (_) {
        stopPlayer().then((_) => setState(() {}));
      });
    }

    return _isPlayerInitialized
        ? _buildWhenPlayerIsInizialized()
        : LoadingProgress();
  }

  Widget _buildWhenPlayerIsInizialized() {
    return Container(
      color: Theme.of(context).colorScheme.surface,
      child: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        crossAxisAlignment: CrossAxisAlignment.center,
        children: [_buildAudioTrackPlayInfo(), _buildActionButtons()],
      ),
    );
  }

  Widget _buildAudioTrackPlayInfo() {
    return Column(
      children: [
        Row(
          mainAxisAlignment: MainAxisAlignment.center,
          children: [
            Text("${getFormattedDuration(_positionAudioTrack)} / ", textScaleFactor: _textScaleSize,),
            Text(getFormattedDuration(_maxAudioTrackDuration), textScaleFactor: _textScaleSize,),
          ],
        ),
        Slider(
            activeColor: Theme.of(context).colorScheme.primary,
            inactiveColor: Theme.of(context).selectedRowColor,
            value: _positionAudioTrackSlider,
            onChanged: (value) {
              var duration = value * _maxAudioTrackDuration.inSeconds;
              _audioPlayer.seekToPlayer(Duration(seconds: duration.toInt()));
            })
      ],
    );
  }

  String _twoDigits(int number) => number.toString().padLeft(2, "0");

  String getFormattedDuration(Duration duration) {
    var hours = _twoDigits(duration.inHours.remainder(24));
    var minutes = _twoDigits(duration.inMinutes.remainder(60));
    var seconds = _twoDigits(duration.inSeconds.remainder(60));
    // var milliseconds = duration.inMilliseconds;

    return "$hours:$minutes:$seconds";
  }

  Widget _buildActionButtons() {
    return ActionButton(
      icon: _audioPlayer.isPlaying
          ? LayoutConfig.audioAndVideoPauseIcon
          : LayoutConfig.audioAndVideoPlayIcon,
      label: _audioPlayer.isPlaying 
        ? S.of(context).audioAndVideoRecordDataPresentationPauseButtonLabel 
        : S.of(context).audioAndVideoRecordDataPresentationPlayButtonLabel,
      onPressed: () => togglePlayAndPause(),
    );
  }

  void togglePlayAndPause() {
    if (!_isPlayerInitialized) {
      return;
    }

    _audioPlayer.isPlaying ? pause() : playOrResume();
  }

  @override
  void dispose() {
    stopPlayer();
    // Be careful : you must `close` the audio session when you have finished with it.
    _audioPlayer.closeAudioSession();
    _audioPlayer = null;

    super.dispose();
  }

  void playOrResume() async {
    if (_audioPlayer.isPaused) {
      _audioPlayer.resumePlayer();
    } else {
      await play();
      setState(() {});
    }
  }

  Future<Duration> play() async{
     return     await _audioPlayer.startPlayer(
          fromURI: widget._suRecordDataFile,
          codec: Codec.aacADTS,
          whenFinished: () {
            stopPlayer();
            setState(() {});
          });
  }

  void pause() async {
    await _audioPlayer.pausePlayer();
  }

  Future<void> stopPlayer() async {
    if (_audioPlayer != null) {
      await _audioPlayer.stopPlayer();
    }
  }
}
