/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:path_provider/path_provider.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_enums.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/util/blocsurecord_data_facade.dart';
import 'package:surec_app/ui/views/record_data_presentation/audio_record_data_presentation.dart';
import 'package:surec_app/ui/views/record_data_presentation/photo_record_data_presentation.dart';
import 'package:surec_app/ui/views/record_data_presentation/note_record_data_presentation.dart';
import 'package:surec_app/ui/views/record_data_presentation/video_record_data_presentation.dart';

enum PresentationMode { PREVIEW, NORMAL }

enum PresentationType {
  VIDEO,
  AUDIO,
  PHOTO,
  NOTE,
}

class RecordDataPresentationUtil{

  static Future<String> loadBlocSuRecordDataForPreview(BlocSuRecord blocSuRecord) async{    
    return await _loadData(blocSuRecord, true);
  }

  static Future<String> loadBlocSuRecordDataForNormalView(BlocSuRecord blocSuRecord) async{    
    return await _loadData(blocSuRecord, false);
  }

  static Widget buildBlocSuRecordTypeIcon(BlocSuRecord blocSuRecord){
    return buildBlocSuRecordPresentation(null, blocSuRecord, true);
  }

  static Widget buildBlocSuRecordPresentation(String data, BlocSuRecord blocSuRecord, bool preview){
    var type = _mapBlocSuRecordTypeToPresentationType(blocSuRecord);

    return data != null
      ? _buildBlocSuRecordDataPresentation(data, type, preview)
      : _buildSuRecordDefaultPresentation(type);
  }

  static Future<String> _loadData(BlocSuRecord blocSuRecord, bool preview) async{
    Directory tempDir = await getTemporaryDirectory();
    String tempPath = tempDir.path;

    var type = _mapBlocSuRecordTypeToPresentationType(blocSuRecord);
    switch(type){
      case PresentationType.NOTE:  
        var noteDataFacade = NoteBlocSuRecordDataFacade(null, tempPath);
        var dataAsText = await noteDataFacade.dataForBlocSuRecord(blocSuRecord, preview);
        return dataAsText;           
      case PresentationType.PHOTO:
        var photoDataFacade = ImageBlocSuRecordDataFacade(null, tempPath);
        var dataAsFilePath = await photoDataFacade.dataForBlocSuRecord(blocSuRecord, preview);
        return dataAsFilePath;
      case PresentationType.VIDEO:   
        var videoDataFacade = VideoBlocSuRecordDataFacade(null, tempPath);
        var dataAsFilePath = await videoDataFacade.dataForBlocSuRecord(blocSuRecord, preview);
        return dataAsFilePath;
      case PresentationType.AUDIO:
        var audioDataFacade = AudioBlocSuRecordDataFacade(null, tempPath);
        var dataAsFilePath = await audioDataFacade.dataForBlocSuRecord(blocSuRecord, preview);
        return dataAsFilePath;
      default:
        throw Exception("type ${type.toString()} is not supported");
    }        
  }

  static Widget _buildBlocSuRecordDataPresentation(String data, PresentationType type, bool preview){
    
    switch(type){
      case PresentationType.NOTE:  
        return NoteRecordDataPresentation(
          suRecordDataAsText: data, 
          presentationMode: preview ? PresentationMode.PREVIEW : PresentationMode.NORMAL
        );
      case PresentationType.PHOTO:  
        return PhotoRecordDataPresentation(
          suRecordDataAsFilePath: data, 
          presentationMode: preview ? PresentationMode.PREVIEW : PresentationMode.NORMAL
        );
      case PresentationType.AUDIO: 
        return preview 
          ? _buildSuRecordDefaultPresentation(type)
          : AudioRecordDataPresentation(suRecordDataAsFilePath: data);
      case PresentationType.VIDEO: 
        return VideoRecordDataPresentation(
          suRecordDataAsFilePath: data, 
          presentationMode: preview ? PresentationMode.PREVIEW : PresentationMode.NORMAL
        );
      default:
        throw Exception("type ${type.toString()} is not supported");
    }
  }

  static Widget _buildSuRecordDefaultPresentation(PresentationType type){
    switch(type){
      case PresentationType.NOTE:  
        return _buildDefaultPresentation(LayoutConfig.noteCaptureRecordIcon);
      case PresentationType.PHOTO:
        return _buildDefaultPresentation(LayoutConfig.photoCaptureRecordIcon);
      case PresentationType.AUDIO:
        return _buildDefaultPresentation(LayoutConfig.audioCaptureRecordIcon);     
      case PresentationType.VIDEO:
        return _buildDefaultPresentation(LayoutConfig.videoCaptureRecordIcon);
      default:
        throw Exception("type ${type.toString()} is not supported");
    }
  }

  static Widget _buildDefaultPresentation(IconData iconData){
    return Center(child: Container(child: Icon(iconData)));
  }

  static PresentationType _mapBlocSuRecordTypeToPresentationType(BlocSuRecord blocSuRecord){
      switch(blocSuRecord.type){
      case SuRecordType.txt:
      case SuRecordType.md:  
        return PresentationType.NOTE;
      case SuRecordType.jpg:
      case SuRecordType.png:
        return PresentationType.PHOTO;
      case SuRecordType.mp3:
      case SuRecordType.wav:
      case SuRecordType.aac:
        return PresentationType.AUDIO;
      case SuRecordType.mp4:
        return PresentationType.VIDEO;
      default:
        throw Exception("blocSuRecord type ${blocSuRecord.type} is not supported");
    }
  }
}