/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';

import 'package:flutter/material.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/views/record_data_presentation/record_data_presentation_util.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:video_player/video_player.dart';
import 'package:surec_app/generated/l10n.dart';

class VideoRecordDataPresentation extends StatefulWidget {
  final File _filePath; 
  final PresentationMode _presentaionMode;

  VideoRecordDataPresentation({
      @required String suRecordDataAsFilePath,
      @required PresentationMode presentationMode})
      : _filePath = File(suRecordDataAsFilePath),
        _presentaionMode = presentationMode;

  @override
  _VideoRecordDataPresentationState createState() =>
      _VideoRecordDataPresentationState();
}

class _VideoRecordDataPresentationState
    extends State<VideoRecordDataPresentation> {
  bool _isVideoPlaying = false;
  VideoPlayerController _videoPlayerController;

  @override
  void initState() {
    super.initState();
    if (widget._presentaionMode == PresentationMode.NORMAL) {
      _videoPlayerController = VideoPlayerController.file(widget._filePath)
        ..initialize().then((_) {
          setState(() {});
        });
    }
  }

  @override
  Widget build(BuildContext context) {
    // return Image.memory(_imageData);
    return widget._presentaionMode == PresentationMode.NORMAL
        ? _buildNormalPresentation()
        : _buildPreviewPresentation();
  }

  Widget _buildPreviewPresentation() {
    return Container(
            decoration: BoxDecoration(
              borderRadius: LayoutConfig.commonBorderRadius,
              image: DecorationImage(
                fit: BoxFit.cover,
                image: FileImage(widget._filePath),
              ),
            ),
          );
  }

  Widget _buildNormalPresentation() {
    return Container(
      color: Theme.of(context).colorScheme.surface,
      padding: const EdgeInsets.only(top: LayoutConfig.simpleWidgetSpacing),
      child: _videoPlayerController.value.isInitialized
          ? Column(
              children: [
                Expanded(
                    child: AspectRatio(
                      aspectRatio: _videoPlayerController.value.aspectRatio,
                      child: VideoPlayer(_videoPlayerController),
                    ),
                  ),
                _buildVideoControlls(),
              ],
            )
          : Container(child: Text(S.of(context).loadingProgressDefaultMessage),),
    );
  }

  Widget _buildVideoControlls() {
    var controllItemColor = Theme.of(context).colorScheme.primary;
    return Container(
      padding: EdgeInsets.all(LayoutConfig.simpleWidgetSpacing),
      child: Row(
        children: [
          ActionButton(
              label: _isVideoPlaying
              ? S.of(context).audioAndVideoRecordDataPresentationPauseButtonLabel
              : S.of(context).audioAndVideoRecordDataPresentationPlayButtonLabel,
              icon: _isVideoPlaying
                  ? LayoutConfig.audioAndVideoPauseIcon
                  : LayoutConfig.audioAndVideoPlayIcon,
              onPressed: () {
                _isVideoPlaying
                    ? _videoPlayerController.pause()
                    : _videoPlayerController.play();

                setState(() {
                  _isVideoPlaying = !_isVideoPlaying;
                });
              }),
          Expanded(
            child: VideoProgressIndicator(              
              _videoPlayerController,
              allowScrubbing: true,
              padding: EdgeInsets.all(LayoutConfig.simpleWidgetSpacing),
              colors: VideoProgressColors(
                playedColor: controllItemColor,
                ),
            ),
          ),
        ],
      ),
    );
  }

  @override
  void dispose() {
    super.dispose();
    if (_videoPlayerController != null) {
      _videoPlayerController.dispose();
    }
  }
}
