/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:flutter_bloc/flutter_bloc.dart';

import 'package:flutter/material.dart';
import 'package:flutter_slidable/flutter_slidable.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/ui/layout_config.dart';
import 'package:surec_app/ui/util/date_time_formatter_util.dart';
import 'package:surec_app/ui/views/detail/read_only_tags.dart';
import 'package:surec_app/ui/views/record_data_presentation/record_data_presentation_util.dart';
import 'package:surec_app/ui/views/record_list/record_export_list_cubit.dart';
import 'package:surec_app/generated/l10n.dart';

class ExpandRecordListItem extends StatefulWidget {
  final BlocSuRecord _blocSuRecord;

  final Function(String id) _onTapItem;
  final Function(BlocSuRecord id) _onDeleteItem;
  final bool _useRecordToExportListAction;
  final RecordExportListCubit _recordExportListCubit;

  ExpandRecordListItem(
      {@required BlocSuRecord blocSuRecord,
      @required Function(String id) onTapItem,
      @required Function(BlocSuRecord blocSuRecord) onDeleteItem,
      bool useRecordToExportListAcion = false,
      RecordExportListCubit recordExportListCubit,
      Key key})
      : this._blocSuRecord = blocSuRecord,
        this._onTapItem = onTapItem,
        this._onDeleteItem = onDeleteItem,
        this._useRecordToExportListAction = useRecordToExportListAcion,
        this._recordExportListCubit = recordExportListCubit ?? null,
        super(key: key);

  @override
  _ExpandRecordListItem createState() => _ExpandRecordListItem();
}

class _ExpandRecordListItem extends State<ExpandRecordListItem> {
  final int _blocSuRecordIdMaxCharacterLength = 10;
  final double _tagPreviewWidthScreenFactor = 0.6;
  final double _blocSuRecordPreviewScreenHeightFactor = 0.3;

  Future<String> _blocSuRecordDataAsFuture;

  Color _borderPreviewColor;
  Color _previewBackgroundColor;
  Color _slideBackgroundColor;

  @override
  void initState() {
    super.initState();

    if (widget._useRecordToExportListAction) {
      assert(widget._recordExportListCubit != null);
    }
    // the record data have to be loaded before the record item is build.
    // When the data is loaded as a child widget in the Slidable widget, the child widget is disposed
    // every time when the list item is swiped left, right or centert.
    // I tried it with keys to keep the child widget alive, but this does not solve the problem.
    // I guess it has something to do with the implementatio of the AutomaticKeepAliveClientMixin in the slideable widget, but I am not sure.
    _blocSuRecordDataAsFuture =
        RecordDataPresentationUtil.loadBlocSuRecordDataForNormalView(
            widget._blocSuRecord);
  }

  @override
  Widget build(BuildContext context) {
    _borderPreviewColor = Theme.of(context).colorScheme.primaryVariant;
    _slideBackgroundColor = Theme.of(context).colorScheme.primaryVariant;
    _previewBackgroundColor = Theme.of(context).colorScheme.background;

    return FutureBuilder(
        future: _blocSuRecordDataAsFuture,
        builder: (BuildContext context, AsyncSnapshot<String> snapshot) {
          Widget widget;
          if (snapshot.hasData) {
            widget = _buildRecordItemWithDataPreview(snapshot.data);
          } else {
            // if data not loaded or error
            widget = _buildRecordItemWithoutDataPreview();
          }

          return widget;
        });
  }

  Widget _buildRecordItemWithoutDataPreview() {
    return _buildRecordItem(null);
  }

  Widget _buildRecordItemWithDataPreview(String suRecordData) {
    return _buildRecordItem(suRecordData);
  }

  Widget _buildRecordItem(String suRecordData) {
    return GestureDetector(
      onTap: () => widget._onTapItem(widget._blocSuRecord.id),
      child: Slidable(
        closeOnScroll: false,
          child: ExpansionTile(             
            title: Row(
              mainAxisAlignment: MainAxisAlignment.spaceBetween,
              children: [
                Container(
                  alignment: Alignment.centerLeft,
                  // padding: const EdgeInsets.only(right: LayoutConfig.doubleWidgetSpacing),
                  child: RecordDataPresentationUtil.buildBlocSuRecordTypeIcon(widget._blocSuRecord),
                ),
                Column(
                  mainAxisAlignment: MainAxisAlignment.center,
                  crossAxisAlignment: CrossAxisAlignment.start,
                  children: [
                    _builDateText(),
                    _buildTags(widget._blocSuRecord.tags),
                  ],
                ),
              ],
            ),
            children: [
              _buildPreview(suRecordData),
            ],
        ),
        actionPane: SlidableDrawerActionPane(),
        actions: widget._useRecordToExportListAction
            ? [
                BlocBuilder<RecordExportListCubit, RecordExportListState>(
                  builder: (context, recordExportListState) {
                    var blocSuRrecord = widget._blocSuRecord;
                    var isAddedToExportList = recordExportListState
                        .recordExportList
                        .contains(blocSuRrecord);

                    return IconSlideAction(
                        closeOnTap: false,
                        caption: isAddedToExportList
                            ? S.of(context).expandRecordListItemIconSliderActionDeleteFromExportLabel
                            : S.of(context).expandRecordListItemIconSliderActionAddToExportLabel,
                        color: _slideBackgroundColor,
                        icon: isAddedToExportList ? Icons.close : Icons.add,
                        onTap: () => isAddedToExportList
                            ? widget._recordExportListCubit
                                .deleteRecordFromExportList(blocSuRrecord)
                            : widget._recordExportListCubit
                                .addRecordToExportList(blocSuRrecord));
                  },
                ),
              ]
            : [],
        secondaryActions: widget._onDeleteItem != null
            ? [
                IconSlideAction(
                  caption: S.of(context).expandRecordListItemIconSliderActionDeleteRecordLabel,
                  color: _slideBackgroundColor,
                  icon: Icons.delete_forever,
                  onTap: () => widget._onDeleteItem(widget._blocSuRecord),
                )
              ]
            : [],
      ),
    );
  }

  Widget _builDateText() {
    var text =
        DateTimeFormatterUtil.formatBlocSuRecordTimeStampToDayMonthYearTime(
            widget._blocSuRecord);
    return Text(text);
  }

  Widget _buildTags(List<String> tags) {
    var tagWidgets = tags
        .map((tag) => Chip(
            backgroundColor: ReadOnlyTags.isReadOnly(tag)
                ? Theme.of(context).colorScheme.secondary
                : Theme.of(context).colorScheme.primaryVariant,
            label: ReadOnlyTags.isReadOnly(tag)
                ? Text(ReadOnlyTags.getTagWithoutPrefix(tag))
                : Text(tag)))
        .toList();

    return SizedBox(
      width: MediaQuery.of(context).size.width * _tagPreviewWidthScreenFactor,
      child: Wrap(
          spacing: LayoutConfig.simpleWidgetSpacing,
          runSpacing: LayoutConfig.simpleWidgetSpacing,
          children: tagWidgets),
    );
  }

  Widget _buildPreview(String suRecordData) {
    return Container(
        padding: EdgeInsets.all(LayoutConfig.doubleWidgetSpacing),
        decoration: BoxDecoration(
          color: _previewBackgroundColor,
          border: Border.all(color: _borderPreviewColor),
          borderRadius: LayoutConfig.commonBorderRadius,
        ),
        height: MediaQuery.of(context).size.height * _blocSuRecordPreviewScreenHeightFactor,
        // width: _dataPreviewSize,
        child: RecordDataPresentationUtil.buildBlocSuRecordPresentation(
            suRecordData, widget._blocSuRecord, false));
  }

  Widget _buildShortHashText() {
    return Text(
        "Hash: ${widget._blocSuRecord.id.substring(0, _blocSuRecordIdMaxCharacterLength)}...");
  }
}
