/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'dart:io';

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:path_provider/path_provider.dart';
import 'package:surec_app/surecord/surecord_blocmeta.dart';
import 'package:surec_app/surecord/surecord_meta.dart';
import 'package:surec_app/surecord/surecord_repo.dart';
import 'package:surec_app/ui/util/export.dart';
import 'package:surec_app/ui/views/record_list/record_export_list_cubit.dart';
import 'package:surec_app/ui/views/record_list/record_list_builder.dart';
import 'package:surec_app/ui/widgets/action_button.dart';
import 'package:surec_app/ui/widgets/card_container.dart';
import 'package:surec_app/generated/l10n.dart';
import 'package:surec_app/ui/widgets/loading_progress.dart';

class RecordExportListView extends StatefulWidget {
  @override
  _RecordExportListViewState createState() => _RecordExportListViewState();
}

class _RecordExportListViewState extends State<RecordExportListView> {
  bool _isExporting = false;

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<RecordExportListCubit, RecordExportListState>(
        builder: (context, recordExportListState) {
      List<BlocSuRecord> blocSuRecords = recordExportListState.recordExportList;
      return blocSuRecords.length > 0
          ? _isExporting
              ? LoadingProgress()
              : Column(
                  children: [
                    _buildExportList(blocSuRecords),
                    _buildExportAction(blocSuRecords),
                  ],
                )
          : _buildEmptyList();
    });
  }

  Widget _buildEmptyList() {
    return Center(
      child: CardContainer(
          child: Text(S.of(context).exportRecordListViewListIsEmptyLabel)),
    );
  }

  Widget _buildExportList(List<BlocSuRecord> blocSuRecords) {
    return Expanded(
      child: CardContainer(
        child: RecordListBuilder(
          blocSuRecords: blocSuRecords,
          useDateRecordDivider: false,
          onDeleteRecord: (blocSuRecord) =>
              BlocProvider.of<RecordExportListCubit>(context)
                  .deleteRecordFromExportList(blocSuRecord),
        ),
      ),
    );
  }

  Widget _buildExportAction(List<BlocSuRecord> blocSuRecords) {
    return ActionButton(
        icon: Icons.import_export,
        label: S.of(context).exportRecordListViewExportButtonLabel,
        onPressed: _isExporting
            ? () {}
            : () {
                _createPdfExport(blocSuRecords).then((_) {
                  setState(() {
                    _isExporting = false;
                  });
                });
                setState(() {
                  _isExporting = true;
                });
              });
  }

  Future<void> _createPdfExport(List<BlocSuRecord> blocSuRecords) async {
    var suRecordFiles = await _getSuRecordFiles(blocSuRecords);
    var pathFileName = await _getPathAndFilename();
    PdfExporter pdfExporter = PdfExporter(suRecordFiles, pathFileName);
    await pdfExporter.writePdf(suRecordFiles, pathFileName);
  }

  Future<List<SuRecordFile>> _getSuRecordFiles(
      List<BlocSuRecord> blocSuRecords) async {
    List<SuRecordFile> suRecordFiles = [];

    SuRecordFilesRepo suRecordFilesRepo = SuRecordFilesRepo();
    for (var blocSuRecord in blocSuRecords) {
      var file = await suRecordFilesRepo.findByID(blocSuRecord.id);
      if (file != null) {
        suRecordFiles.add(file);
      }
    }

    return suRecordFiles;
  }

  Future<String> _getPathAndFilename() async {
    Directory documentDirectory = await getApplicationDocumentsDirectory();
    String documentPath = documentDirectory.path;
    return Future.value("$documentPath/example.pdf");
  }
}
