/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec_app - prototype fund round 8 release.

    surec_app is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec_app is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:surec_app/surecord/surecord_cubit.dart';
import 'package:surec_app/ui/views/record_list/record_export_list_cubit.dart';
import 'package:surec_app/ui/views/record_list/record_list_builder.dart';
import 'package:surec_app/ui/views/record_list/record_list_cubit.dart';
import 'package:surec_app/ui/widgets/card_container.dart';
import 'package:surec_app/ui/widgets/filter_options/date_filter.dart';
import 'package:surec_app/ui/widgets/filter_options/filter_option_cubit.dart';
import 'package:surec_app/ui/widgets/filter_options/search_text_filter.dart';
import 'package:surec_app/ui/widgets/filter_options/tag_filter.dart';
import 'package:surec_app/generated/l10n.dart';

class RecordListView extends StatelessWidget {
  final RecordExportListCubit _recordExportListCubit;

  RecordListView(this._recordExportListCubit);

  @override
  Widget build(BuildContext context) {
    var suRecordCubit = BlocProvider.of<SuRecordCubit>(context);
    return MultiBlocProvider(
      providers: [
        BlocProvider<RecordListCubit>(
            create: (context) =>
                RecordListCubit(suRecordCubit)
                  ..initialRecordList()),
        BlocProvider<FilterOptionCubit>(
            create: (context) => FilterOptionCubit()),
      ],
      child: BlocConsumer<RecordListCubit, RecordListState>(
        listener: (context, listState) {
          // record list error handling
        },
        builder: (context, listState) {
          var recordListCubit = BlocProvider.of<RecordListCubit>(context);
          var filterOptionCubit = BlocProvider.of<FilterOptionCubit>(context);
          var records = listState.filteredBlocSuRecords;

          return BlocListener<FilterOptionCubit, FilteOptionState>(
              listener: (context, filterOptionState) {
                if (filterOptionState is UpdateFilterOptionState) {
                  var filterExpression =
                      filterOptionState.filterExpressions.values.toList();
                  recordListCubit.filterBlocSuRecords(filterExpression);
                }
              },
              child: Column(
                mainAxisAlignment: MainAxisAlignment.start,
                mainAxisSize: MainAxisSize.max,
                children: [
                  _buildFilterPanel(suRecordCubit, filterOptionCubit),
                  records.length > 0
                      ? Expanded(
                          child: CardContainer(
                            child: RecordListBuilder(
                              blocSuRecords: records,
                              onDeleteRecord: (blocSuRecord) {
                                recordListCubit
                                    .deleteBlocSuRecord(blocSuRecord);
                                _recordExportListCubit
                                    .deleteRecordFromExportList(blocSuRecord);
                              },
                              useRecordToExportListAction: true,
                              useDateRecordDivider: true,
                            ),
                          ),
                        )
                      : Expanded(
                          child: Center(
                            child: CardContainer(
                              child: Text(
                                  S.of(context).recordListViewListIsEmptyLabel),
                            ),
                          ),
                      ),
                ],
              ));
        },
      ),
    );
  }

  Widget _buildFilterPanel(SuRecordCubit suRecordCubit, FilterOptionCubit filterOptionCubit) {
    return CardContainer(
      child: ExpansionTile(
        title: SearchTextFilter(),
        children: [
          _buildTagFilterOption(suRecordCubit, filterOptionCubit),
          _buildDateFilterOption(filterOptionCubit),
        ],
      ),
    );
  }

  Widget _buildDateFilterOption(FilterOptionCubit cubit) {
    return CardContainer(child: DateFilter(cubit));
  }

  Widget _buildTagFilterOption(SuRecordCubit suRecordCubit, FilterOptionCubit filterOptionCubit) {
    return CardContainer(child: TagFilter(suRecordCubit, filterOptionCubit, filterMode: TagFilterMode.SUBTRACTIONAL,));
  }
}
