/*
    Copyright (c) 2021 Richard Meinsen.

    This file is part of surec backend - prototype fund round 8 release.

    surec backend is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    surec backend is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.surec.backend.controller

import eu.surec.backend.repositories.ConfirmationRepository
import eu.surec.backend.repositories.ConfirmationEntity
import eu.surec.backend.repositories.FingerprintEntity
import eu.surec.backend.dto.Algo
import eu.surec.backend.dto.Confirmation
import eu.surec.backend.dto.Fingerprints
import eu.surec.backend.util.toHex
import org.slf4j.LoggerFactory
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.web.bind.annotation.*
import java.nio.charset.StandardCharsets
import java.security.MessageDigest


@RestController
@RequestMapping("/api/v1")
class ConfirmationController(@Autowired private val confirmationRepository : ConfirmationRepository) {
    private val log = LoggerFactory.getLogger(ConfirmationController::class.java)

    @GetMapping("/", )
    fun showConfig() : ConfirmationControllerConfig {
        log.info("get/showConfig", "Returning \$(ConfirmationWebserviceConfig())")
        return ConfirmationControllerConfig();
    }

    @GetMapping("/findByConfirmationValueAndAlgo")
    fun findByConfirmationValueAndAlgo(@RequestParam value : String, @RequestParam algo : Algo) : List<Confirmation> {
        log.info("get/findByConfirmationValueAndAlgo", "value $value  , algo $algo")

        var confirmationEntities = confirmationRepository.findByValueAndAlgo(value, algo);
        var confirmations = mutableListOf<Confirmation>()
        for(confirmationEntity in confirmationEntities) {
            try {
                var confirmationAlgo: Algo = confirmationEntity.algo as Algo
                confirmations.add(Confirmation(confirmationEntity.value, confirmationAlgo, confirmationEntity.timestamp))
            } catch (e : Exception) {
                log.error("get/findByConfirmationValueAndAlgo", "algo $algo invalid")
            }
        }
        return confirmations;
    }

    @GetMapping("/findByConfirmationValue")
    fun findByConfirmationValue(@RequestParam value : String) : List<Confirmation> {
        log.info("get/findByConfirmationValue", "value $value")

        var confirmationEntities = confirmationRepository.findByValue(value);
        var confirmations = mutableListOf<Confirmation>()
        for(confirmationEntity in confirmationEntities) {
            try {
                var algo: Algo = confirmationEntity.algo as Algo
                confirmations.add(Confirmation(confirmationEntity.value, algo, confirmationEntity.timestamp))
            } catch (e : Exception) {
                log.error("get/findByConfirmationValue", "value $value invalid")
            }
        }
        return confirmations;
    }

    @GetMapping("/findAll")
    fun findAll() : List<Confirmation> {
        log.info("get/findAll", "")

        var confirmationEntities = confirmationRepository.findAll();
        var confirmations = mutableListOf<Confirmation>()
        for(confirmationEntity in confirmationEntities) {
            try {
                var algo: Algo = confirmationEntity.algo as Algo
                confirmations.add(Confirmation(confirmationEntity.value, algo, confirmationEntity.timestamp))
            } catch (e : Exception) {
                log.error("get/findAll", e)
            }
        }
        return confirmations;
    }

    @PostMapping("/confirm")
    fun confirm(@RequestBody fingerprints : Fingerprints): Confirmation {
        log.info("get/confirm", "")

        val timestamp = System.currentTimeMillis() // https://www.codevscolor.com/kotlin-current-time-milliseconds

        val messageDigest: MessageDigest = MessageDigest.getInstance(ConfirmationControllerConfig().confirmationValueAlgo.value)

        var confirmationBase: String = ""
        var fingerprintEntities = mutableListOf<FingerprintEntity>()

        for(fp in fingerprints.fingerprints) {
            log.info("got fingerprint ${fp.algo} - ${fp.value}")
            confirmationBase += fp.value
            val fpe : FingerprintEntity = FingerprintEntity(0, fp.algo, fp.value)
            fingerprintEntities.add(fpe)
        }

        log.info("get/confirm", "confirmationbase : $confirmationBase")
        val confirmationBytes  = messageDigest.digest(confirmationBase.toByteArray(StandardCharsets.UTF_8))
        val confirmationHex = confirmationBytes.toHex()
        log.info("get/confirm", "confirmation : $confirmationHex")

        var confirmationEntity : ConfirmationEntity = ConfirmationEntity(
                0,confirmationHex,ConfirmationControllerConfig().confirmationValueAlgo, timestamp, fingerprintEntities)
        confirmationRepository.save(confirmationEntity)

        val confirmation = Confirmation(confirmationHex, ConfirmationControllerConfig().confirmationValueAlgo, timestamp)

        return confirmation
    }

}